package uaa.feral.context;

import gov.nasa.worldwind.geom.LatLon;
import gov.nasa.worldwind.geom.Position;
import gov.nasa.worldwind.render.Polyline;
import gov.nasa.worldwind.render.Renderable;
import gov.nasa.worldwind.render.SurfacePolygon;
import gov.nasa.worldwind.render.UserFacingIcon;

import java.awt.Color;
import java.awt.Dimension;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import repast.simphony.context.Context;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.gis.Geography;
import repast.simphony.space.graph.Network;

import cern.jet.random.Uniform;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Point;

import uaa.feral.physical.HydrologicalSetting;
import uaa.feral.physical.HydrologicalUnit;
import uaa.feral.physical.behaviors.HydroBehavior;
import uaa.feral.physical.behaviors.RemoteSensingEstimator;
import uaa.feral.social.Community;
import uaa.feral.social.Organization;
import uaa.feral.social.Person;
import uaa.feral.social.demographics.DemographicFunctions;
import uaa.feral.social.demographics.Sex;
import uaa.feral.social.demographics.StandardDemographicTable;
import uaa.feral.social.leadership.AbstractLeadershipModel;
import uaa.feral.social.leadership.LeadershipModel;
import uaa.feral.social.waterUse.WaterUseBehaviors;
import uaa.feral.social.waterUse.WaterUseModel;
import uaa.feral.util.data.ScenarioData;
import uaa.feral.util.loader.CSVLoader;
import uaa.feral.util.loader.ShapeFileLoader;

/**
 * Class for creating a scenario for a simulation based on input files and xml data.
 * @author 
 */

public class ScenarioBuilder {

	
	private int hydroNumber;  //number to keep track of hydrological systems
//	private List<String>waterSourceNames;
	private HashMap<String, Map<Integer, double[]>> hydroSourceData;  //hydrological source data
	private List<Community> communities;  //list of communities in this scenario
	private SocialNetworkBuilder snb;
	
	public ScenarioBuilder(Network population,Context context){
		hydroNumber=0;
	//	waterSourceNames=new ArrayList<String>();
		Person.Id=0;
		snb=new SocialNetworkBuilder(population,context);
	}
	
	/**
	 * Method to build a population from initial scenario information
	 * @param sData the scenario data
	 * @param waterUseModel the water use model
	 * @param geography the geography object
	 * @return a list of communities
	 */
	public List<Community> buildPopulation(ScenarioData sData,Geography geography){
		List<Community> communities = new ArrayList<Community>();
		//load the file here
		if(sData.getCommunityFile().contains(".shp")){
			List<Map> communityData=ShapeFileLoader.ShapeFileLoader(new 
					File(FeralContextBuilder.PathToData+sData.getCommunityFile()));
			Iterator<Map> ic = communityData.iterator();
			int i=0;
			while(ic.hasNext()){
				Map<String,Object> data = ic.next();
				String name = (String)data.get("NAME");
				if(!name.equals(sData.getName()))
					continue;
				Geometry geo = (Geometry)data.get("the_geom");
				Community community = new Community(name,(Point)geo);
				List<Person>people=null;
				if(sData.getWaterUseModel()!=null)
					people=createPeople(data,community);
				else if(sData.getLeadershipModel()!=null)
					people=createNetworkPopulation(sData,community);
				community.addAll(people);
				addCommunityDetails(community,sData);
				setupSocialModels(community,sData);
				geography.move(community, geo);
				communities.add(community);
				i++;
			}
		}
		this.communities=communities;
		return communities;
	}
	
	private List<Person> createNetworkPopulation(ScenarioData sd, Community community){
		List<Map>data=CSVLoader.loadCSVStatic(FeralContextBuilder.PathToData+sd.getSocialNetworkFile());
		List people=snb.parseSocialNetworkData(data,community);
		return people;
	}
	
	private void setupSocialModels(Community community, ScenarioData sd){
		if(sd.getWaterUseModel()!=null) {
			WaterUseModel wModel=WaterUseBehaviors.CreateWaterUseModel(sd.getWaterUseModel(),sd,community);
			community.setWaterUseModel(wModel);
		}
		if(sd.getLeadershipModel()!=null){
			LeadershipModel leadershipModel=AbstractLeadershipModel.CreateLeadershipModel(sd.getLeadershipModel(),sd,community);
			community.setLeadershipModel(leadershipModel);
		}
	}
		
	/**
	 * Method to create a population for a town based on file data
	 * @param data file data 
	 * @param community the community object
	 * @return a list of people
	 */
	public List<Person> createPeople(Map<String,Object> data,Community community){
		List<Person>people = new ArrayList<Person>();
		long id=0;
		Geometry geo = (Geometry)data.get("the_geom");
		Point point=(Point)geo;
		Set<String> keys = data.keySet();
		Iterator<String> ik = keys.iterator();
		while(ik.hasNext()){
			String key = ik.next();
			int popCategory=0;
			try{
				popCategory=((Double)data.get(key)).intValue();
			}
			catch(ClassCastException e){
				
			}
			String[] split=null;
			if(key.startsWith("MA")|| key.startsWith("FE")) {
				Sex sex =null;
				if(key.startsWith("MA"))
						sex=Sex.MALE;
				else
					sex=Sex.FEMALE;
				if(key.contains("_"))
					split=key.split("_");
				if(split==null){
					if(key.endsWith("OVER")){
						List<Person> cPeople=createPeopleSpread(community,popCategory,sex,point,85,100);
						people.addAll(cPeople);
						id+=cPeople.size();
						continue;
					}
					for(int i=0; i < popCategory;i++){
						Person person = new Person(community, Integer.parseInt(key.substring(key.length()-1)),sex,point);
						people.add(person);
					}
				}
				else {
					if(split.length==1){
						for(int i=0; i < popCategory;i++){
							int age=Integer.parseInt(split[0].substring(split[0].length()-1))-1;
							Person person = new Person(community,age,sex,point);
							people.add(person);
						}
					}
					else{
						int ageFrom=Integer.parseInt(split[0].substring(split[0].length()-2));
						int ageTo=Integer.parseInt(split[1].substring(split[1].length()-2));
						List<Person> cPeople=createPeopleSpread(community,popCategory,sex,point,ageFrom,ageTo);
						people.addAll(cPeople);
					}
				}	
			}
		}
		return people;
	} 
	
	/**
	 * Method to create people based on population data
	 * @param community the person's community
	 * @param number 
	 * @param sex sex type
	 * @param point the location
	 * @param ageFrom beginning age
	 * @param ageTo end age
	 * @return
	 */
	public List<Person> createPeopleSpread(Community community,double number,Sex sex,Point point,int ageFrom,int ageTo){
		List<Person>people = new ArrayList<Person>();
		Uniform uniform=RandomHelper.createUniform();
		
		for(int i=0; i < number; i++){
			int age=uniform.nextIntFromTo(ageFrom, ageTo);
			Person person = new Person(community,age,sex,point);
			people.add(person);
		}
		
		return people;
	}
	
	/**
	 * Method to load file data describing community characteristics
	 * @param community the community
	 * @param data the data for the scenario
	 */
	public void addCommunityDetails(Community community,ScenarioData data){
		double fertilityRate=0d;
		double migrationRate=0d;
		List<Map>communityData=CSVLoader.loadCSVStatic(FeralContextBuilder.PathToData+data.getDemographicFile());
		Iterator<Map> ic = communityData.iterator();
		double[] rates = new double[101];
		while(ic.hasNext()){
			Map<String,Object> dataPoint = ic.next();
			if(!((String)dataPoint.get("Geography")).trim().equals(community.getId()))
				continue;
			Iterator<String> si = dataPoint.keySet().iterator();
			while(si.hasNext()){
				String key = si.next();
				if(key.startsWith("Death Rate")){
					double rate = Double.valueOf((String)dataPoint.get(key));
					String[] split = key.split(" ");
					String last = split[3];
					if(last.contains("-")) {
						String[] splitLast=last.split("-");
						int value1 = Integer.parseInt(splitLast[0]);
						int value2 = Integer.parseInt(splitLast[1]);
						rates=createRates(value1,value2,rates,rate,100000);
					}
					else if(last.contains("+")){
						String modified=last.substring(0, last.length()-1);
						int value=Integer.parseInt(modified);
						rates=createRates(value,100,rates,rate,100000);
					}
				}
			}
			fertilityRate=(Double)Double.valueOf((String)dataPoint.get("Birth Rate"));
			migrationRate=(Double)Double.valueOf((String)dataPoint.get("Migration"));
		}
		
		if(data.isDemographics()) {
			StandardDemographicTable sdt = new StandardDemographicTable(rates);
			DemographicFunctions df = new DemographicFunctions(community,fertilityRate,migrationRate);
			df.setTable(sdt);
			community.setDemographicFunction(df);
		}
	}
	
	/**
	 * Function for creating demographic rates
	 * @param value1 start value
	 * @param value2 end value
	 * @param rates an array of rates for the values
	 * @param rate the numerator to divide and get the rate
	 * @param denom the denominator used for the rate
	 * @return
	 */
	private double[] createRates(int value1, int value2, double[] rates,double rate,double denom){
		for(int i=value1; i<=value2;i++){
			rates[i]=rate/denom;
		}
		return rates;
	}
	
	/**
	 * Method to build hydrology for the scenario
	 * @param sd the scenario data containing data references
	 * @param hydrologicalModel the hydrology model to use
	 * @param geography the geography context
	 * @return
	 */
	public HydrologicalSetting buildHydrology(ScenarioData sd, String hydrologicalModel,Geography geography){
		
		HydrologicalSetting hs = new HydrologicalSetting(sd,hydroNumber,hydrologicalModel);
		Map<String,Map<Integer,double[]>> dischargeData=null;
		
		//setup discharge info
		if(sd.getStreamFlowFile()!=null) {
			dischargeData=createDischargeData(sd.getStreamFlowFile(),hs);
		}
		
		//load the file here
		if(sd.getHydrologicalFile()!=null && sd.getHydrologicalFile().contains(".shp")){
			List<Map>hydroData=ShapeFileLoader.ShapeFileLoader(new File(
					FeralContextBuilder.PathToData+sd.getHydrologicalFile()));
			Iterator<Map> ih = hydroData.iterator();
			
			while(ih.hasNext()){
				Map<String,Object>data = ih.next();
				Geometry geo = (Geometry)data.get("the_geom");
				String nameR = (String)data.get("GNIS_Name");
				if(dischargeData!=null && !dischargeData.containsKey(nameR))
					continue;
				ArrayList latLongs = new ArrayList();
				Coordinate[] coords=geo.getCoordinates();
				double size=1;
				Renderable shape=null;
				if(nameR.contains("River") || nameR.contains("Creek")) {
					size=1;
					if(nameR.contains("River"))
						size=2;
					for(int i=0; i<coords.length;i++){
						Coordinate c = coords[i];
						latLongs.add(Position.fromDegrees(c.y, c.x, 0));
					}
					shape = new Polyline(latLongs);
					((Polyline) shape).setAntiAliasHint(Polyline.ANTIALIAS_NICEST);
					((Polyline) shape).setColor(Color.BLUE);	
					((Polyline) shape).setLineWidth(size);
					((Polyline) shape).setPathType(Polyline.LINEAR);
					((Polyline) shape).setFollowTerrain(true);
				}
				else{
					for(int i=0; i<coords.length;i++){
						Coordinate c = coords[i];
						latLongs.add(LatLon.fromDegrees(c.y, c.x));
					}					
					shape=new SurfacePolygon(latLongs,Color.BLUE,Color.BLUE);
				}
				HydrologicalUnit hu = new HydrologicalUnit(geo,shape);
				hu.setName(nameR);
				if(hs.getHydroBehavior().getName().equals("RemoteSensingEstimator"))
					buildRemoteSensingHydroModel(hs,hu,data);
				hs.add(hu);
			}	
		}
		if(sd.isMunicipalSystem()){
			buildMunicipalSystem(hs,sd);
		}
		hydroNumber++;
		
		return hs;
	}
	
	public Map<String,Map<Integer,double[]>> createDischargeData(String file,HydrologicalSetting hs){
		List<Map>riverFlowData=CSVLoader.loadCSVStatic(FeralContextBuilder.PathToData+file);
		Map<String, Map<Integer, double[]>> dischargeData = returnRiverDischargeData(riverFlowData);
		HydroBehavior bh=hs.getHydroBehavior();
		Iterator<String>ki = dischargeData.keySet().iterator();
		while(ki.hasNext()){
			String key = ki.next();
			Map<Integer,double[]> localData = dischargeData.get(key);
			Iterator<Integer>kii = localData.keySet().iterator();
			while(kii.hasNext()){
				int k = kii.next();
				bh.setDischargeInfo(key,k, localData);
			}
		}	
		return dischargeData;
	}
	
	/**
	 * Method to initialize remote sensing-based hydrological discharge model.
	 * @param hs the Hydrological setting
	 * @param hu the hydrological unit
	 * @param data the input data
	 */
	public void buildRemoteSensingHydroModel(HydrologicalSetting hs, HydrologicalUnit hu, Map<String,Object>data){
		Iterator<String> ic = data.keySet().iterator();
		
		double slope=((Double)data.get("slope")).doubleValue();
		double bankfullWidth=((Double)data.get("bankfullWidth")).doubleValue();
		
		hu.setSlope(slope);
		hu.setBankfullWidth(bankfullWidth);
		
		while(ic.hasNext()){
			String key = ic.next();
			if(key.contains("waterWidth")){
				String[]month=key.split("idth");
				double result=Double.parseDouble((String)data.get(key));
				Map<HydrologicalUnit,Map<Integer,Double>> wValues=
					((RemoteSensingEstimator)hs.getHydroBehavior()).getHydroloigcalWidth();
				
				Map<Integer, Double> iData=null;
				if(!wValues.containsKey(key)) 
					iData = new HashMap<Integer,Double>();
				else
					iData=wValues.get(hu);	
				
				iData.put(Integer.parseInt(month[1]), result);
				wValues.put(hu, iData);
			}
		}
		((RemoteSensingEstimator)hs.getHydroBehavior()).setHydrologicalSetting(hs);
	}
	
	/**
	 * Method for creating a municipal water system based on input data
	 * @param hs a hydrological setting object to add hydrological unit(s) to
	 */
	private void buildMunicipalSystem(HydrologicalSetting hs,ScenarioData sd){
		Iterator<Community> ic = communities.iterator();
		while(ic.hasNext()){
			Community c = ic.next();
			Map<String,Map<Integer,double[]>>dData=createDischargeData(sd.getMunicipalFile(),hs);
			Map<Integer,double[]>dataM=dData.get(c.getId());
			HydrologicalUnit hu = new HydrologicalUnit(c.getLocation(),null);
			hu.setMunicipalSource(true);
			hu.setName((String)c.getId()+" Municipal");
			dData.put(hu.getName(), dataM);
			dData.remove(c.getId());
			hs.getHydroBehavior().setData(dData);
			hs.add(hu);
		}
	}
	
	/**
	 * Method to get discharge data from file
	 * @param riverFlowData the discharge data loaded from a file
	 * @return a map with organized discharge data based on monthly discharge
	 */
	public Map<String,Map<Integer,double[]>> returnRiverDischargeData(List<Map>riverFlowData){
		
		if(hydroSourceData==null)
			hydroSourceData = new HashMap<String,Map<Integer,double[]>>();
		Map<String,Map>tempData = new HashMap<String,Map>();
		Iterator<Map> ri = riverFlowData.iterator();
		while(ri.hasNext()){
			Map data = ri.next();
			Map<Integer,List>temp = null;
			List<Double> monthData=null;
			double meanV =Double.parseDouble((String) data.get("mean_va"));
			int month = Integer.parseInt((String)data.get("month_nu"))-1;
			if(!hydroSourceData.containsKey(data.get("Water Body"))){
					temp=new HashMap<Integer,List>();
					tempData.put((String) data.get("Water Body"), temp);
					hydroSourceData.put((String) data.get("Water Body"), null);
			}
			temp=tempData.get((String)data.get("Water Body"));
			if(temp.containsKey(month)){
				monthData = temp.get(month);
				monthData.add(meanV);
			}
			else{
				monthData=new ArrayList<Double>();
				monthData.add(meanV);
				temp.put(month, monthData);
			}
		}
		
		Iterator<String> ik=tempData.keySet().iterator();
		while(ik.hasNext()){
			String river = ik.next();
			Map<Integer,List>temps=tempData.get(river);
			Iterator<Integer>ikk=temps.keySet().iterator();
			Map<Integer,double[]> disch=new HashMap<Integer,double[]>();
			while(ikk.hasNext()){
				int key = ikk.next();
				List<Double>mData = temps.get(key);
				double[] data = new double[mData.size()];
				for(int i=0; i < mData.size(); i++){
					data[i]=mData.get(i);
				}
				disch.put(key, data);
			}
			hydroSourceData.put(river, disch);
		}
		return hydroSourceData;
	}	
	
	/**
	 * Method to create Organizations
	 * @param sd scenario data input
	 * @param model the model to associate with water use in the mine
	 * @param geography the geography the mine belongs to
	 * @return a list of organizations
	 */
	public List<Organization> buildMine(ScenarioData sd, String model){
		List<Organization> organizations = new ArrayList<Organization>();
		if(sd.getMiningFile().contains(".shp")){
			List<Map> miningData=ShapeFileLoader.ShapeFileLoader(new 
					File(FeralContextBuilder.PathToData+sd.getMiningFile()));
			Iterator<Map> ic = miningData.iterator();
			
			while(ic.hasNext()){
				Map<String,Object> data = ic.next();
				String name = (String)data.get("Name");
				if(!name.equals(sd.getName()))
					continue;
				Geometry geo = (Geometry)data.get("the_geom");
				Organization o = new Organization(name,(Point)geo);
				UserFacingIcon icon = new UserFacingIcon("../FERAL/lib/mining.png",
		                  Position.fromDegrees(geo.getCentroid().getY(), geo.getCentroid().getX(),0d));
		        icon.setSize(new Dimension(44, 44));
		        o.setIcon(icon);    
		       
		        WaterUseModel wModel=WaterUseBehaviors.CreateWaterUseModel(model,sd,o);
				o.setWaterUseModel(wModel);
				
				organizations.add(o);
			}
		}
		return organizations;
	}
}
