package uaa.feral.context;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import repast.simphony.context.Context;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.graph.Network;

import uaa.feral.social.Community;
import uaa.feral.social.Person;
import uaa.feral.social.demographics.Sex;
import uaa.feral.social.network.AgentType;
import uaa.feral.social.network.RelationshipEdge;
import uaa.feral.social.network.Status;

/**
 * Class for creating social networks in simulations 
 * with existing data.
 * 
 * @author 
 */
public class SocialNetworkBuilder {

	private HashMap<Integer,Person> numberedAgents;
	private Map<AgentType,Integer> agentTypes;
	private Network population;
	private Context context;

	/**
	 * Default constructor
	 * @param population the social network 
	 * @param context the simulation context to add the population
	 */
	public SocialNetworkBuilder(Network population, Context context){
		this.numberedAgents=new HashMap<Integer,Person>();
		this.population=population;
		this.context=context;
		
		this.agentTypes=new HashMap<AgentType,Integer>();
		this.agentTypes.put(AgentType.Alpha, 0);
		this.agentTypes.put(AgentType.Beta, 1);
		this.agentTypes.put(AgentType.Gamma, 2);		
	}
	
	/**
	 * Method to parse the initial social network input
	 * @param data list of data points
	 * @param community the community associated with individuals
	 * @return a list of person agents
	 */
	public List<Person> parseSocialNetworkData(List<Map> data,Community community){
		List<Person>people=new ArrayList<Person>();
		
		Iterator<Map> ic = data.iterator();
		Map<Person,Map<String,String>>tempData=new HashMap<Person,Map<String,String>>();
		
		while(ic.hasNext()){
			Map<String,Object>dataPoint = ic.next();
			Map<String,String>tempPoint=new HashMap<String,String>();
			
			String type=(String) dataPoint.get("Type");
			AgentType at=null;
			if(type.equalsIgnoreCase("beta"))
				at=AgentType.Beta;
			else if(type.equalsIgnoreCase("alpha"))
				at=AgentType.Alpha;
			else if(type.equalsIgnoreCase("gamma"))
				at=AgentType.Gamma;
			
			String importance = (String)dataPoint.get("Importance");
			String cost = (String)dataPoint.get("DecisionCost");
			Person p = new Person(community,1,Sex.MALE,community.getLocation());
			p.setAgentType(at);
			p.setImportance(Double.valueOf(importance));
			p.setDecisionCost(Double.valueOf(cost));
			
			String weightA=(String) dataPoint.get("WeightA");
			String weightB=(String) dataPoint.get("WeightB");
			String weightC=(String) dataPoint.get("WeightC");
			
			tempPoint.put("WeightA",weightA);
			tempPoint.put("WeightB",weightB);
			tempPoint.put("WeightC",weightC);
			tempData.put(p, tempPoint);
			numberedAgents.put(p.getId(),p);
			
			people.add(p);
			if(!context.contains(p))
				context.add(p);
		}
		createNetwork(tempData);
		return people;
	}
	
	/**
	 * Method that creates the social network
	 * @param tempData a map with people and their network data
	 */
	public void createNetwork(Map<Person,Map<String,String>>tempData){
		Iterator<Person> ic = tempData.keySet().iterator();
		int size=0;
		int n=0;
		while(ic.hasNext()){
			Person p = ic.next();
			n++;
			numberedAgents.remove((int)p.getId());
			Map<String,String>tempPoint = tempData.get(p);
			
			String wA=tempPoint.get("WeightA");
			String wB=tempPoint.get("WeightB");
			String wC=tempPoint.get("WeightC");
			
			String[] weights={wA,wB,wC};
			
			double[] mV=new double[weights.length];
			double[] mE=new double[weights.length];
			
			for(int i=0; i <weights.length; i++){
				String[] wS=((String)weights[i]).split("/");
				mV[i]=Double.parseDouble(wS[0]);
				mE[i]=Double.parseDouble(wS[1]);
			}
			
			Iterator<Person> in = numberedAgents.values().iterator();
			while(in.hasNext()){
				Person np = in.next();
				if(p==np)
					continue;
				int agentType=this.agentTypes.get(np.getAgentType());
				double beta=RandomHelper.createBeta
					(mV[agentType],mE[agentType]).nextDouble();
				double wF=(beta-.5)*2;
			//	System.out.println(wF);
				createLink(p,np,wF);
				size++;
			}
			numberedAgents.put((int)p.getId(), p);
		}
	}
	
	/**
	 * Method to create the social links in the network
	 * @param p1 person 1 (the source)
	 * @param p2 person 2 (the target)
	 * @param weightF the weight of the edge
	 */
	public void createLink(Person p1, Person p2,double weightF){
		RelationshipEdge re = new 
			RelationshipEdge(p1,p2,p2.getAgentType());
		re.setSourceTargetIds(p1.getId(), p2.getId());
		re.setWeight(weightF);
		population.addEdge(re);
		context.add(re);
		
	}
}
