package uaa.feral.physical;

import gov.nasa.worldwind.awt.WorldWindowGLCanvas;
import gov.nasa.worldwind.layers.RenderableLayer;
import gov.nasa.worldwind.render.Polyline;
import gov.nasa.worldwind.render.Renderable;

import java.awt.Color;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import repast.simphony.context.DefaultContext;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import repast.simphony.engine.schedule.IAction;
import repast.simphony.engine.schedule.ISchedule;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.parameter.Parameters;
import repast.simphony.query.PropertyEquals;
import repast.simphony.space.gis.Geography;
import repast.simphony.visualization.IDisplay;
import repast.simphony.visualization.gis3D.DisplayGIS3D;
import uaa.feral.physical.behaviors.EvolveHydrology;
import uaa.feral.physical.behaviors.HydroBehavior;
import uaa.feral.style.WWSurfaceCircleFERAL;
import uaa.feral.util.data.ScenarioData;

import gov.nasa.worldwind.geom.LatLon;

/**
 * Class to contain the hydrological unit(s) in a simulation
 * @author 
 *
 * @param <HydrologicalUnit>
 */
public class HydrologicalSetting extends DefaultContext<HydrologicalUnit>{

	public HydroBehavior hydroBehavior;  //physical model governing hydrological changes/evolution
	private Map<HydrologicalUnit,Double>priorAmount;  //record of amounts
	private DisplayGIS3D gis;  //3D visual display
	private double riverScale;  //scale factor used for river systems display
	private double municipalScale; //scale factor used for municipal systems display
	/**
	 * Constructor that can be used for limited hydrological behaviors or generic hydrological modeling for a scenario
	 */
	public HydrologicalSetting(){
		super("hydroSetting","Hydro Setting");
		this.riverScale=Double.parseDouble((String)RunEnvironment.getInstance().getParameters().getValue("waterScaleFactor"));
		this.municipalScale=Double.parseDouble((String)RunEnvironment.getInstance().getParameters().getValue("municipalScaleFactor"));
	}

	/**
	 * Constructor for testing
	 * @param test a string to indicate a test
	 */
	public HydrologicalSetting(String test){
		super(test,test);
	}
	
	/**
	 * Main constructor for a hydrological setting
	 * @param number the number of the setting
	 * @param hydroModel the hydrology model associated with the setting
	 */
	public HydrologicalSetting(ScenarioData sd,int number,String hydroModel){
		super(Integer.toString(number),"Hydro Setting");
		priorAmount=new HashMap<HydrologicalUnit,Double>();
		hydroBehavior=EvolveHydrology.CreateHydroModel(sd,hydroModel);
		this.riverScale=Double.parseDouble((String)RunEnvironment.getInstance().getParameters().getValue("riverScaleFactor"));
		this.municipalScale=Double.parseDouble((String)RunEnvironment.getInstance().getParameters().getValue("municipalScaleFactor"));
	}
	
	/**
	 * Scheduled class for running hydrology
	 * @author Mark Altaweel
	 */
	private class RunHydrologics implements IAction{
		public void execute() {
			runHydro();
		}
	}
	
	/**
	 * Method to initialize the object prior to execution of model behaviors
	 */
	@ScheduledMethod(start = 0, pick=ScheduledMethod.ALL, priority = Double.POSITIVE_INFINITY)
	public void init(){
		 Parameters p = RunEnvironment.getInstance().getParameters(); 
		 Geography geography = (Geography)RunState.getInstance().getMasterContext().getProjection("Geography");
//		 int daysPerTick=(Integer)p.getValue("daysPerTick");
//		 double ticks = (double) 365.25/daysPerTick;
		 
//		 if(ticks>12)
//			 ticks=ticks/12;
//		 else
//			 ticks=1;
		 
		 ISchedule schedule = repast.simphony.engine.environment.RunEnvironment.getInstance()
	      	.getCurrentSchedule();
		  schedule.schedule(ScheduleParameters.createRepeating(1, 1,Double.POSITIVE_INFINITY),new RunHydrologics());
		 
		 List<IDisplay> displays=RunState.getInstance().getGUIRegistry().getDisplays();
		 Iterator<IDisplay> ic = displays.iterator();
		 while(ic.hasNext()){
			 IDisplay display = ic.next();
			 if(display instanceof DisplayGIS3D){
				 if(gis==null)
					 gis = (DisplayGIS3D)display;
				 WorldWindowGLCanvas canvas=gis.getWwglCanvas();
				 Iterator<HydrologicalUnit> hui=this.getObjects(HydrologicalUnit.class).iterator();
				 while(hui.hasNext()){
					 HydrologicalUnit hu = hui.next();
					 Renderable shape = hu.getRenderable();
					 if(shape==null) {
						 shape=new WWSurfaceCircleFERAL(gis.getWwglCanvas().getModel().getGlobe(),LatLon.fromDegrees(hu.getGeometry().getCentroid().getY(),
								 hu.getGeometry().getCentroid().getX()),100,100,Color.BLUE,Color.BLUE);
						 hu.setRenderable(shape);
					 }
					 RenderableLayer rl = new RenderableLayer();
					 rl.addRenderable(shape);
					 canvas.getModel().getLayers().add(rl);
	//				 canvas.redraw();		 
				 }
			 }
		 }
	}
	
	/**
	 * Method to run hydrological model or calculation 
	 */
	public void runHydro(){
		hydroBehavior.runModel();
		Map<String,Double>waterMap=hydroBehavior.getWaterQuantityByBody();
		Iterator<String> ic = waterMap.keySet().iterator();
		while(ic.hasNext()){
			 String river = ic.next();
			 PropertyEquals<HydrologicalUnit> hus = new PropertyEquals<HydrologicalUnit>(this,
					  "name", river);
			 
			 Iterator<HydrologicalUnit> hi=hus.query().iterator();
			 while(hi.hasNext()){
				 HydrologicalUnit hu = hi.next(); 
				 double waterQ=waterMap.get(river);
				 if(!priorAmount.containsKey(hu))
					 priorAmount.put(hu,waterQ);
				 if(!hu.isMunicipalSource()) {
					 hu.setDischarge(waterQ);
					 hu.calculateDailyVolumeFromDischarge();
				 }
				 else 
					 hu.setVolume(waterQ);
				 setWaterColorScale(hu);
	//			 break;  //probably should get rid of this and it should iterate through all units, for now leave as is until data changes
			 }
		}
	}
	
	/**
	 * Method to set the colors in WW
	 * @param hu a hydrological unit
	 */
	public void setWaterColorScale(HydrologicalUnit hu){
		double high=priorAmount.get(hu);
		double v=hu.getQuality();
		double scale=0;
		//if no volume or discharge, then we are assuming this is a hydrological quality sim only
		if(hu.getVolume()>0 || hu.getDischarge() > 0) {
			if(hu.getName().endsWith("Municipal")) {
				v=hu.getVolume();
				scale=this.municipalScale;
			}
			else {
				v=hu.getDischarge();
				scale=this.riverScale;
			}
		}
		
		Color c=Color.BLUE;
		if(v/high < 1-scale) 
			c=Color.CYAN;
		if(v/high < 1-(scale*2))
			c=Color.GREEN;
		if(v/high < 1-(scale*3))
			c=Color.MAGENTA;
		if(v/high < 1-(scale*4))
			c=Color.ORANGE;
		if(v/high < 1-(scale*5))
			c = Color.RED;
		if(v>high)
			priorAmount.put(hu, v);   
		 
		if(hu.getRenderable() instanceof Polyline)
			 ((Polyline)hu.getRenderable()).setColor(c);
		 else if(hu.getRenderable() instanceof WWSurfaceCircleFERAL){
			((WWSurfaceCircleFERAL)hu.getRenderable()).setPaint(c);
		 }
	}

	public HydroBehavior getHydroBehavior() {
		return hydroBehavior;
	}

	public void setHydroBehavior(HydroBehavior hydroBehavior) {
		this.hydroBehavior = hydroBehavior;
	}
}
