package uaa.feral.physical.behaviors;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.parameter.Parameters;
import repast.simphony.random.RandomHelper;

import uaa.feral.physical.HydrologicalSetting;
import uaa.feral.physical.HydrologicalUnit;

/**
 * Discharge estimated using remote sensing information.
 * 
 * 
 * Model developed using Bjerklie et. al. 2003 article in Journal Hydrology 278:17-38
 * See also Bjerklie 2007 Journal Hydrology 341:144-155 for Velocity calculation
 * @author 
 *
 */
public class RemoteSensingEstimator extends EvolveHydrology{

//	private Map<HydrologicalUnit,Double>gravityForRiver=new HashMap<HydrologicalUnit,Double>();

	/**
	 * Default constructor
	 */
	public RemoteSensingEstimator(){
		super();
		Parameters p = RunEnvironment.getInstance().getParameters();
		daysPerTick=(Double)p.getValue("daysPerTick");
		name="RemoteSensingEstimator";
	}
	
	/**
	 * Used for testing the model
	 * @param string a test string
	 */
	public RemoteSensingEstimator(String string) {
		super();
	}

	private HydrologicalSetting hs;
	
	private Map<HydrologicalUnit,Map<Integer,Double>> hydrologicalWidth=new HashMap<HydrologicalUnit,Map<Integer,Double>>();

	/**
	 * Rate of acceleration due to gravity
	 * @param a the latitude location
	 * @return a gravity acceleration value in meters (e.g. 9.822 m/s squared)
	 */
	public double determineGravity(double a){
		double g=0;
		g=(9.7703267714)*((1+(0.00193185138639*Math.pow(Math.sin(a),2)))
				/Math.sqrt(1-0.00669437999013*Math.pow(Math.sin(a),2)));
		
		return g;
	} 
	
	/**
	 * Method used to determine the length fraction of a meander. 
	 * @param hu a hydrological unit
	 * @return the length fraction
	 */
	public double determineLengthFraction(HydrologicalUnit hu){
		double slope=hu.getSlope();
		double meanderLength=hu.getAverageMeanderLength();
		if(meanderLength<=0) {
			meanderLength=10.2*Math.pow(hu.getBankfullWidth(), 1.12);
			hu.setAverageMeanderLength(meanderLength);
		}
		double m=9.67*Math.pow(slope*meanderLength,.36);
		
		return m;
	} 
	
	/**
	 * Method used to determine velocity using g (gravity pull) and m (meander length fraction).
	 * @param hu the hydroloigcal unit
	 * @param g gravity acceleration
	 * @param m meander length fraction
	 * @return a velocity value
	 */
	public double factorVelocity(HydrologicalUnit hu,double g,double m){
		double numerator=2*g*hu.getAverageMeanderLength()*hu.getSlope();
		
		return Math.sqrt(numerator/m);
	}
	
	/**
	 * Set up current width value of the river
	 */
	public void setupWidthValue(HydrologicalUnit hu,int month){
		Map<Integer,Double>wValues=hydrologicalWidth.get(hu);	
		
		double width=wValues.get(month);
		
		double currentWidth=RandomHelper.createPoisson(width).nextDouble();
		hu.setWaterWidth(currentWidth);
	}
	
	/**
	 * Method to determine river velocity. Note:  Method not currently used.
	 * @param hu the hydrological unit
	 * @return a double value velocity
	 */
	public double determineVelocity(HydrologicalUnit hu){
		double v=0d;
		
		double s=hu.getSlope();
		double l=hu.getAverageMeanderLength();
		
		v=1.37*Math.pow(s, .31)*Math.pow(l, .32);
		return v;
	}
	
	/**
	 * Method for determining discharge using determined velocity
	 * @param hu the hydrological unit modeled
	 * @param velocity the velocity of the given hydrological unit
	 * @return a discharge value
	 */
	public double determineDischarge(HydrologicalUnit hu,double velocity){
		double q=0d;
		
		q=0.09*Math.pow(hu.getWaterWidth(),1.21)*Math.pow(velocity, 1.53)*Math.pow(hu.getSlope(),-.3);
		
		return q;
	}

	public void setDischargeInfo(String river, int month,
			Map<Integer, double[]> sample) {
		// TODO Auto-generated method stub
	}
	
	
	public void runModel(){
		Iterator<HydrologicalUnit>hi=hs.iterator();
		
		int month=getCalendarMonth();
		
		while(hi.hasNext()){
			HydrologicalUnit hu = hi.next();
			setupWidthValue(hu,month);
			double g=determineGravity(hu.getGeometry().getCentroid().getY());
			double m=determineLengthFraction(hu);
			double v=factorVelocity(hu,g,m);
	//		double v=determineVelocity(hu);
			double discharge=determineDischarge(hu,v);
			hu.setDischarge(discharge);  //discharge is set for individual hydrological units
		}
	}
	
	/**
	 * For testing the run method. 
	 * @param month a month used for testing
	 */
	public void runModel(int month){
		Iterator<HydrologicalUnit>hi=hs.iterator();
		
		while(hi.hasNext()){
			HydrologicalUnit hu = hi.next();
			setupWidthValue(hu,month);
			double g=determineGravity(hu.getGeometry().getCentroid().getY());
			double m=determineLengthFraction(hu);
			double v=factorVelocity(hu,g,m);
	//		double v=determineVelocity(hu);
			double discharge=determineDischarge(hu,v);
			hu.setDischarge(discharge);  //discharge is set for individual hydrological units
		}
	}

	public HydrologicalSetting getHydrologicalSetting() {
		return hs;
	}

	public void setHydrologicalSetting(HydrologicalSetting hs) {
		this.hs = hs;
	}

	public Map<HydrologicalUnit, Map<Integer, Double>> getHydroloigcalWidth() {
		return hydrologicalWidth;
	}

	public void setHydroloigcalWidth(Map<HydrologicalUnit, Map<Integer, Double>> hydrologicalWidth) {
		this.hydrologicalWidth = hydrologicalWidth;
	}
}
