package uaa.feral.social;

import gov.nasa.worldwind.geom.Position;
import gov.nasa.worldwind.layers.IconLayer;
import gov.nasa.worldwind.render.Renderable;
import gov.nasa.worldwind.render.UserFacingIcon;

import java.awt.Dimension;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import com.vividsolutions.jts.geom.Point;

import repast.simphony.context.DefaultContext;
import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import repast.simphony.engine.schedule.IAction;
import repast.simphony.engine.schedule.ISchedule;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.engine.schedule.ScheduledMethod;
import repast.simphony.parameter.Parameters;
import repast.simphony.visualization.IDisplay;
import repast.simphony.visualization.gis3D.AbstractDisplayLayerGIS3D;
import repast.simphony.visualization.gis3D.DisplayGIS3D;
import repast.simphony.visualization.gis3D.SphereShape;
import uaa.feral.physical.HydrologicalUnit;
import uaa.feral.social.demographics.DemographicFunctions;
import uaa.feral.social.leadership.LeadershipModel;
import uaa.feral.social.waterUse.WaterUseModel;

/**
 * Community object that contains person agents.
 * The community does display agent behavior, in addition to containing person agents
 * @author
 *
 * @param <Person>
 */
public class Community extends DefaultContext<Person>  {

	WaterUseModel waterUseModel;  //model for water use
	LeadershipModel leadershipModel;  //for leaders, followers, and opportunists type scenarios
	Point location;  //location of the community
	HydrologicalUnit municipalSupply;  //municipal water supply
	DemographicFunctions demographicFunction;  //behaviors governing demographic changes
	private SphereShape shape;  //shape that governs visual object's size
	private UserFacingIcon icon;  //visual icon of the community
	private double daysPerTick;
	private double tickYearly;
	private double importance;
	
	public double getImportance() {
		return importance;
	}

	public void setImportance(double importance) {
		this.importance = importance;
	}

	public Community(){
		 super("id","Community");
	}
	
	/**
	 * Main constructor used to create a community
	 * @param id the community's id
	 * @param location the community's location
	 */
	public Community(String id,Point location){
		super(id,"Community");
		this.location=location;
		this.daysPerTick=0;
		this.tickYearly=0;
//		if(RunEnvironment.getInstance().getCurrentSchedule().getTickCount()>0)
//			init();
	}
	
	/**
	 * Class to run scheduled demographic functions
	 * @author
	 *
	 */
	private class RunDemographics implements IAction{
		public void execute() {
			runDemographics();
		}
	}
	
	/**
	 * Class to run scheduled population manager in demographic functions
	 * @author
	 *
	 */
	private class ManageDemographics implements IAction{
		public void execute() {
			managePopulation();
		}
	}
	
	/**
	 * Class to run scheduled social model behaviors
	 * @author 
	 *
	 */
	private class RunModels implements IAction{
		public void execute(){
			runModels();
		}
	}
	
	public HydrologicalUnit getMunicipalSupply() {
		return municipalSupply;
	}

	public Point getLocation() {
		return location;
	}

	public void setLocation(Point location) {
		this.location = location;
	}

	public void setMunicipalSupply(HydrologicalUnit municipalSupply) {
		this.municipalSupply = municipalSupply;
	}

	/**
	 * Method to initialize parameters, behaviors, and call visualization initialization
	 */
	@ScheduledMethod(start = 0, pick=ScheduledMethod.ALL, priority = Double.MAX_VALUE)
	public void init(){
		Parameters p = RunEnvironment.getInstance().getParameters(); 
		
		 if(daysPerTick!=0 || this.tickYearly!=0)
			 return;
		 this.daysPerTick=Double.parseDouble((String)p.getValue("daysPerTick"));
		 this.tickYearly = (double) 365.25/daysPerTick;
		 
		  if(this.demographicFunction!=null) {
			  ISchedule schedule = RunEnvironment.getInstance()
		      	.getCurrentSchedule();
			  schedule.schedule(ScheduleParameters.createRepeating(1, tickYearly,2),new RunDemographics());
			  
			  ISchedule schedule1 = RunEnvironment.getInstance()
		      	.getCurrentSchedule();
			  schedule1.schedule(ScheduleParameters.createRepeating(1, tickYearly,1),new ManageDemographics());	
		  }
		  
		  if(tickYearly>12)
				 tickYearly=tickYearly/12;
			 else
				 tickYearly=1;
		
		 
		  ISchedule schedule2 = repast.simphony.engine.environment.RunEnvironment.getInstance()
		      	.getCurrentSchedule();
		  schedule2.schedule(ScheduleParameters.createOneTime(1, ScheduleParameters.LAST_PRIORITY),new RunModels());
		  
		  if(!RunEnvironment.getInstance().isBatch())
			  initV();	  
	}
	
	/**
	 * Method that initializes a community visual object (in animation mode)
	 */
	private void initV(){
	
			List<IDisplay> displays=RunState.getInstance().getGUIRegistry().getDisplays();
			Iterator<IDisplay> ic = displays.iterator();
			DisplayGIS3D gis=null;
			while(ic.hasNext()){
				IDisplay display = ic.next();
				if(display instanceof DisplayGIS3D)
					gis=(DisplayGIS3D)display;
			}
			if(gis==null)
				return;
			Map<Class,AbstractDisplayLayerGIS3D> objs=gis.getClassStyleMap();
			AbstractDisplayLayerGIS3D layer = objs.get(Community.class);
			Collection<Renderable> renderables=(Collection<Renderable>)layer.getRenderables();
			Iterator<Renderable> ico=layer.getShapeToObjectMap().keySet().iterator();
			while(ico.hasNext()){
				Renderable r = ico.next();
				if(layer.getShapeToObjectMap().get(r) == this){
					this.shape=(SphereShape)r;
					((SphereShape)r).setRadius(-10);
				}
			}
            IconLayer layer2 = new IconLayer();

            UserFacingIcon icon = new UserFacingIcon("../FERAL/img/houses_icon.gif",
                  Position.fromDegrees(shape.getGeometry().getCentroid().getY(), shape.getGeometry().getCentroid().getX(),0));
            icon.setSize(new Dimension(44, 44));
            this.icon=icon;
            layer2.addIcon(icon);
			gis.getWwglCanvas().getModel().getLayers().add(layer2);
//			gis.getWwglCanvas().redrawNow();
			
//			PlaceNameLayer l = new PlaceNameLayer(new PlaceNameServiceSet());
//			gis.getWwglCanvas().getModel().getLayers().add(l);
			if(demographicFunction!=null)
				demographicFunction.setGis(gis);
			
	}
	
	/**
	 * Method to run the water use and/or leadership model(s)
	 */
	public void runModels(){
		if(waterUseModel!=null)
			waterUseModel.runModel();
		
		if(leadershipModel!=null)
			leadershipModel.runModel();
	}
	
	
	/**
	 * Method to run demographics
	 */
	public void runDemographics(){
		demographicFunction.demographicSteps();
	}
	
	/**
	 * Method manages a population using the demographic behaviors created
	 */
	public void managePopulation(){
		demographicFunction.managePopulationSize();
	}
	
	public WaterUseModel getWaterUseModel() {
		return waterUseModel;
	}

	public void setWaterUseModel(WaterUseModel waterUseModel) {
		this.waterUseModel = waterUseModel;
	}

	public DemographicFunctions getDemographicFunction() {
		return demographicFunction;
	}

	public void setDemographicFunction(DemographicFunctions demographicFunction) {
		this.demographicFunction = demographicFunction;
	}
	
	public SphereShape getShape(){
		return shape;
	}

	public UserFacingIcon getIcon() {
		return icon;
	}

	public void setIcon(UserFacingIcon icon) {
		this.icon = icon;
	}

	public LeadershipModel getLeadershipModel() {
		return leadershipModel;
	}

	public void setLeadershipModel(LeadershipModel leadershipModel) {
		this.leadershipModel = leadershipModel;
	}
}
