package uaa.feral.social.demographics;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import repast.simphony.engine.schedule.IAction;
import repast.simphony.engine.schedule.ISchedule;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.parameter.Parameters;
import repast.simphony.random.RandomHelper;
import repast.simphony.visualization.IDisplay;
import repast.simphony.visualization.gis3D.DisplayGIS3D;
import uaa.feral.social.Community;
import uaa.feral.social.Person;

/**
 * Class for conducting basic demographic functions for a person (e.g. birth, death, etc.)
 * @author 
 */
public class DemographicFunctions {

	StandardDemographicTable table;  //mortality mainly
	double fertilityRate;  //rate per 1000
	double migrationRate; //rate per 1000
	Community community;  //community the demographic evolution is relevant for
	int targetSize;
	double modifier=1;
	double townSize=1;
	Double popRateCeiling;
	DisplayGIS3D gis;
	boolean initial=true;
	double initialSize=0;
	Map<Integer,List<Person>>birthMap=new HashMap<Integer,List<Person>>();
	Map<Integer,List<Person>>deathMap=new HashMap<Integer,List<Person>>();
	double tickYearly;
	double townScale;
	
	public DemographicFunctions(){
	}
	
	/**
	 * Main constructor used for creating demographic behaviors
	 * @param community the community that will have its demographics evolve
	 * @param fertilityRate a given fertility rate for a community
	 * @param migrationRate a migration (immigration/emigration) rate for the community
	 */
	public DemographicFunctions(Community community,double fertilityRate, double migrationRate){
		Parameters p = RunEnvironment.getInstance().getParameters(); 
		
		this.community=community;
		this.fertilityRate=(double)fertilityRate/1000;
		this.migrationRate=(double)migrationRate/1000;
		this.popRateCeiling=Double.parseDouble((String)p.getValue("populationTargetSize"));
		 
		double daysPerTick=Double.parseDouble((String)p.getValue("daysPerTick"));
		this.tickYearly = (double) 365.25/daysPerTick;
		this.townScale=Double.parseDouble((String)p.getValue("townSizeScale"));
	}
	
	/**
	 * Class for scheduling dying, birth, and migration
	 * @author Mark Altaweel
	 */
	private class ScheduledPeople implements IAction{
		List<Person> life;
		List<Person> death;
 		public ScheduledPeople(List<Person>life,List<Person>death){
			this.life=life;
			this.death=death;
		}
		
		public void execute() {
			if(death!=null)
				dieOrEmigrate(death);
			if(life!=null)
				bornOrImmigrate(life);		
		}
	}
	
	/**
	 * Method called that launches the demographic methods
	 */
	public void demographicSteps(){
		if(initial) {
			targetSize=(int)(community.size()*popRateCeiling);
			initialSize=community.size();
			initial=false;
		}
		birthMap.clear();
		deathMap.clear();
		age();
		birth();
		migrate();
		if(RunEnvironment.getInstance()!=null)
			scheduleDynamics();
		if(community.getIcon()!=null)
			vizPop();
	}
	
	/**
	 * Schedule when demographic functions (birth, death, and migration) will occur in a year
	 */
	private void scheduleDynamics(){
		double currentTick=RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
		for(int i=0; i < (int)tickYearly;i++){
			List<Person>death=deathMap.get(i);
			List<Person>life=birthMap.get(i);
			
			if(life!=null || death!=null) {
				ISchedule schedule = RunEnvironment.getInstance()
	      		.getCurrentSchedule();
				if(life!=null) {
				
					schedule.schedule(ScheduleParameters.createOneTime(i+currentTick, Double.POSITIVE_INFINITY),new ScheduledPeople(life,null));
				}
				if(death!=null){
					schedule.schedule(ScheduleParameters.createOneTime(i+currentTick, Double.POSITIVE_INFINITY),new ScheduledPeople(null,death));
				}
				
			}
		}
	}
	
	/**
	 * In animation mode this will visualize the community based on its size
	 */
	private void vizPop(){
		double radius=community.getIcon().getSize().getWidth();
		if((double)community.size()/initialSize <= townSize-townScale){ 
			radius-=1;
			townSize-=townScale;
		}
		if((double)community.size()/initialSize >= townSize+townScale) {
			radius+=1;
			townSize+=townScale;
		}
		community.getIcon().getSize().setSize(radius,radius);
	}
	
	 /**
	  * Method created to artificially manage a population's size
	  */
	 public void managePopulationSize(){
		int currentSize=community.size();

	    double[] mort=table.getProbabilities();
	    double r0=0;
	    for(int i=14; i < 85; i++){
	    	r0+=(1-mort[i])*fertilityRate;
	    }
	    double rN=1+(r0-1)*(1-((double)currentSize/targetSize));
	    double multiplier=(rN/r0);
	    
	    modifier=multiplier;
	 }
	 
	 /**
	  * Method that allows people to age, with some of the person agents dying due to probability
	  */
	private void age(){
		List<Person>forDeath = null;
		Iterator<Person>pi=community.getObjects(Person.class).iterator();
		while(pi.hasNext()){
			Person p = pi.next();
			double probDeath=table.demographicFunction(p);
			double random=RandomHelper.nextDouble();
			if(p.getAge()>=90)  //this is needed because we don't have good pop. demographic data for deaths over 85
				probDeath+=(.048*(p.getAge()-89));
			if(random<probDeath) {
				int tick=RandomHelper.nextIntFromTo(0, (int)tickYearly);
				if(deathMap.containsKey(tick)){
					forDeath=deathMap.get(tick);
					forDeath.add(p);
				}
	
				else{
					forDeath=new ArrayList<Person>();
					forDeath.add(p);
					deathMap.put(tick, forDeath);
				}
				removeVisualLayer(p);
				continue;
			}
			p.setAge(p.getAge()+1);
		}
	}
	
	/**
	 * New agents are born into the community
	 */
	private void birth(){
		List<Person>births=null;
		int size = community.size();
		if(this.targetSize<size)
			modifier-=.1;
		double number = size*(fertilityRate*modifier);
		number=Math.round(number);
		for(int i=0; i < number;i++){
			Person p = new Person(community,0,table.determineSex(),community.getLocation());
			int tick=RandomHelper.nextIntFromTo(0,(int)tickYearly);
			if(birthMap.containsKey(tick)){
				births=birthMap.get(tick);
				births.add(p);
			}
			else{
				births=new ArrayList<Person>();
				births.add(p);
				birthMap.put(tick, births);
			}
		}
		
	}
	
	/**
	 * Agents immigrating into a community
	 * @param number the number that will immigrate
	 */
	private void immigrate(int number){
		List<Person>immigrate=null;
		for(int i=0; i < number;i++){
			int age=RandomHelper.nextIntFromTo(0, 80);
			Person p = new Person(community,age,table.determineSex(),community.getLocation());
			int tick=RandomHelper.nextIntFromTo(0,(int)tickYearly);
			if(birthMap.containsKey(tick)){
				immigrate=birthMap.get(tick);
				immigrate.add(p);
			}
			else{
				immigrate=new ArrayList<Person>();
				immigrate.add(p);
				birthMap.put(tick, immigrate);
			}			
		}
		
	}
	
	/**
	 * Agents born or immigrating get added here
	 * @param newPeople the list of agents to add
	 */
	private void bornOrImmigrate(List<Person>newPeople){
		community.addAll(newPeople);
	}
	
	private void removeVisualLayer(Person p){
		if(gis==null){
			List<IDisplay> displays=RunState.getInstance().getGUIRegistry().getDisplays();
			Iterator<IDisplay> ic = displays.iterator();		   
			while(ic.hasNext()){
				   IDisplay display = ic.next();
				   if(display instanceof DisplayGIS3D) {
					   gis=(DisplayGIS3D)display;
				   	   break;
				   }
			   }
		}
		 if(gis!=null){
			 gis.getWwglCanvas().getModel().getLayers().remove(p.getIcon());
			 gis.getWwglCanvas().redraw();
		}		
	}
	/**
	 * Emigrating or dying agents leave the community here and are no longer considered alive
	 * @param deathMigration
	 */
	private void dieOrEmigrate(List<Person>deathMigration){	   
		Iterator<Person>pi=deathMigration.iterator();
		while(pi.hasNext()) {
			Person p = pi.next();
			p.setAlive(false);
		}
		community.removeAll(deathMigration);
	}
	
	/**
	 * Method that migrates agent (immigrate or emmigrate)
	 */
	private void migrate(){
		int size = community.size();
		int number = (int) (size * migrationRate);
		if(number<0){
			List<Person>migration=null;
			for(int i=number; i <0;i++){
				int remove=RandomHelper.createUniform().nextIntFromTo(0, community.size()-1);
				Person p = (Person) community.getObjects(Person.class).get(remove);
				int tick=RandomHelper.nextIntFromTo(0, (int)tickYearly);
				if(deathMap.containsKey(tick)){
					migration=deathMap.get(tick);
					migration.add(p);
				}
				else{
					migration=new ArrayList<Person>();
					migration.add(p);
					deathMap.put(tick, migration);
				}
				removeVisualLayer(p);
			}
		}
		else if(number>0){
			immigrate(number);
		}
	}
	
	public StandardDemographicTable getTable() {
		return table;
	}

	public void setTable(StandardDemographicTable table) {
		this.table = table;
	}

	public double getFertilityRate() {
		return fertilityRate;
	}

	public void setFertilityRate(double fertilityRate) {
		this.fertilityRate = fertilityRate;
	}

	public double getMigrationRate() {
		return migrationRate;
	}

	public void setMigrationRate(double migrationRate) {
		this.migrationRate = migrationRate;
	}

	public Community getCommunity() {
		return community;
	}

	public void setCommunity(Community community) {
		this.community = community;
	}

	public int getTargetSize() {
		return targetSize;
	}

	public void setTargetSize(int targetSize) {
		this.targetSize = targetSize;
	}
	
	public void setGis(DisplayGIS3D gis){
		this.gis=gis;
	}
}
