package uaa.feral.social.leadership;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import org.jmat.data.AbstractMatrix;
import org.jmat.data.Matrix;

import cern.jet.random.Beta;
import cern.jet.random.Uniform;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import repast.simphony.engine.schedule.IAction;
import repast.simphony.engine.schedule.ISchedule;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.parameter.Parameters;
import repast.simphony.random.RandomHelper;
import repast.simphony.space.graph.Network;

import uaa.feral.context.FeralContextBuilder;
import uaa.feral.context.OutputContext;
import uaa.feral.social.Community;
import uaa.feral.social.Person;
import uaa.feral.social.data.AppliedSocialModelData;
import uaa.feral.social.network.AgentType;
import uaa.feral.social.network.RelationshipEdge;
import uaa.feral.util.PCA;

/**
 * Model evaluates agent responses and satisfaction, then adjusts 
 * agent social weights in a network.
 * 
 * Concepts are similar to Kitts (2006):
 * Kitts, James A. 2006. Social influence and the emergence of norms 
 * amid ties of amity and enmity. Simulation Modelling Practice and 
 * Theory 14:407-422.
 * 
 * However, differences are found between this and Kitts' 
 * model. 
 * 
 * @author 
 */
public class AgentTypesModel extends AbstractLeadershipModel{

	private Community community;
	private AppliedSocialModelData modelData;
	private int initialCount=0;
	private Map<Person,Double>normativeDependence;
	private Map<Person,Double>affectiveDependence;
	private Map<Person,Double>complianceBias;
	private Map<RelationshipEdge, Double>approvalBias;
	private Map<Person,Double>propensityToComply;
	private Uniform uniformPropensity;
	private double tickNow;
	private double endTick;
	private double individualShare;
	private PCA pca;
	private OutputContext context;
	
	/**
	 * Class to run scheduled function call. 
	 * @author 
	 */
	private class RunModel implements IAction{
		public void execute(){
			runModel();
		}
	}
	
	/**
	 * Initialize the model and associate it with a community object.
	 * @param community the community object
	 */
	public AgentTypesModel(Community community){
		this.community=community;
		this.affectiveDependence=new HashMap<Person,Double>();
		this.complianceBias=new HashMap<Person,Double>();
		this.approvalBias=new HashMap<RelationshipEdge,Double>();
		this.normativeDependence=new HashMap<Person,Double>();
		this.propensityToComply=new HashMap<Person,Double>();
		this.pca=new PCA();
	}
	
	/**
	 * Method executes the model at each time tick, with the method initially called by the community 
	 * object.
	 */
	public void runModel() {
		this.tickNow=(double)RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
		if(this.context==null)
			this.context=(OutputContext)RunState.getInstance().getMasterContext().getSubContext("output");
		if(initialCount==0)
			setupModelData(scenarioData.getSocialModelDataFile());
		eigenSpace();
		individualBenefitShare();
		Iterator<Person>ic = community.iterator();
		
		while(ic.hasNext()){
			Person p = (Person)ic.next();
			double propensityToComply=propensityToComply(p);
			Iterator<Person> ip = this.affectiveDependence.keySet().iterator();
			while(ip.hasNext()){
				Person person=ip.next();
				if(person==p)
					continue;
				else
					weightChange(p,person,propensityToComply);		
			}
			this.propensityToComply.put(p, propensityToComply);
		}
		
		Iterator<Person> ip = this.propensityToComply.keySet().iterator();
		while(ip.hasNext()){
			Person p = ip.next();
			double value = this.propensityToComply.get(p);
			updateChoicePropensity(p,value);
		}
		
		if(tickNow==endTick){
			eigenSpace();
			context.compileOutput();  //finalize output to csv file(s)
			RunEnvironment.getInstance().getCurrentSchedule().setFinishing(true);  //end simulation processes
		}
		
		//schedule the model to rerun every tick
		ISchedule schedule = repast.simphony.engine.environment.RunEnvironment.getInstance()
      		.getCurrentSchedule();
		schedule.schedule(ScheduleParameters.createOneTime(tickNow+1, Integer.MAX_VALUE),new RunModel());
		initialCount++;  //used as a flag to set initial conditions
	}
	
	/**
	 * Method to setup model data from an input xml file.
	 * @param file the file with xml data
	 */
	public void setupModelData(String file) {
		AppliedSocialModelData  smd = new AppliedSocialModelData();
		this.modelData=smd.setupModelData(FeralContextBuilder.PathToData+file);
		setupInitialScheduleInfo();
		setupRejectAcceptNormativeAffective();
		context.setSocialTitleLine("Tick,"+"Id,"+"X,"+"Y,"+"Comply");
	}
	
	/**
	 * Setup model time length.
	 */
	public void setupInitialScheduleInfo(){
		Parameters p = RunEnvironment.getInstance().getParameters();
		double ticks=Double.parseDouble((String)p.getValue("ticks"));
		this.endTick=ticks;
		RunEnvironment.getInstance().endAt(Double.valueOf(endTick));
	}
	
	/**
	 * More setting up of initial model data derived from xml data.
	 */
	public void setupRejectAcceptNormativeAffective(){
		Beta a = RandomHelper.createBeta(modelData.getStatusWeightedDecisions().get("affectiveDependence1"), 
				modelData.getStatusWeightedDecisions().get("affectiveDependence1E"));
		Beta b = RandomHelper.createBeta(modelData.getStatusWeightedDecisions().get("affectiveDependence2"), 
				modelData.getStatusWeightedDecisions().get("affectiveDependence2E"));
		Beta g = RandomHelper.createBeta(modelData.getStatusWeightedDecisions().get("affectiveDependence3"), 
				modelData.getStatusWeightedDecisions().get("affectiveDependence3E"));

		Beta learningA = RandomHelper.createBeta(modelData.getStatusWeightedDecisions().get("normativeDependence1"), 
				modelData.getStatusWeightedDecisions().get("normativeDependence1E"));
		Beta learningB = RandomHelper.createBeta(modelData.getStatusWeightedDecisions().get("normativeDependence2"), 
				modelData.getStatusWeightedDecisions().get("normativeDependence2E"));
		Beta learningG = RandomHelper.createBeta(modelData.getStatusWeightedDecisions().get("normativeDependence3"), 
				modelData.getStatusWeightedDecisions().get("normativeDependence3E"));
		
		Iterator ic = community.iterator();
		while(ic.hasNext()){
			Object o = ic.next();
			if(o instanceof Person){
				Person p=(Person)o;
				
				if(p.getAgentType()==AgentType.Alpha) {
					this.normativeDependence.put(p, learningA.nextDouble());
					this.affectiveDependence.put(p, a.nextDouble());
					p.setComply(1);
				}
				else if(p.getAgentType()==AgentType.Beta) {
					this.normativeDependence.put(p, learningB.nextDouble());
					this.affectiveDependence.put(p, b.nextDouble());
					if(RandomHelper.nextDouble()<.5)
						p.setComply(1);
					else
						p.setComply(-1);
				}
				else {
					this.normativeDependence.put(p, learningG.nextDouble());
					this.affectiveDependence.put(p, g.nextDouble());
					if(RandomHelper.nextDouble()>p.getDecisionCost())
						p.setComply(1);
					else
						p.setComply(-1);
				}	
				
				Network network = (Network)RunState.getInstance().getMasterContext().getProjection("Population");
		
				Iterator<RelationshipEdge> ri=network.getOutEdges(p).iterator();
			
				while(ri.hasNext()){
					RelationshipEdge re = ri.next();
					double r=RandomHelper.createUniform(-1, 1).nextDouble();
					if(re.getWeight()<-.1)
						re.setApproval(-1);
					else if(re.getWeight()>.1)
						re.setApproval(1);
					else{
						if(re.getWeight()<r)
							re.setApproval(-1);
						else 
							re.setApproval(1);
					}
					
				}
						
			}
		}
	}
	
	/**
	 * Function to determine where, based on a PCA of agent 
	 * weight values, does the agent lie in relation to other 
	 * agents.
	 */
	public void eigenSpace(){
		Network network = (Network)RunState.getInstance().getMasterContext().getProjection("Population");
		int size=this.normativeDependence.keySet().size();
		Matrix m = new Matrix(size,size);
		List<Person> temp = new ArrayList<Person>();
		Iterator<Person> ip = this.normativeDependence.keySet().iterator();
		while(ip.hasNext()){
			Person p = ip.next();
			Iterator<Person>ic = this.affectiveDependence.keySet().iterator();
			while(ic.hasNext()){
				Person person = ic.next();
				if(p==person){
					m.set(p.getId(),p.getId(), 1);
					continue;
				}
				RelationshipEdge<Person> re = (RelationshipEdge) network.getEdge(p, person);
				m.set(p.getId(), person.getId(), re.getWeight());
			}
			temp.add(p);
		}
		pca.runPca(m);
	
		AbstractMatrix vectors=pca.getVectors().transpose();
		AbstractMatrix one=vectors.getRow(temp.size()-2);
		AbstractMatrix two=vectors.getRow(temp.size()-1);

		for(int i=0; i < temp.size();i++){
			Person p=temp.get(i);
			double x = one.get(0, p.getId());
			double y = two.get(0, p.getId());
			p.setX(x);
			p.setY(y);
			context.recordAgentOutput(p,(int)tickNow);  //record agent data here
		}
	}
	
	/**
	 * S=(1/N)*Sum(ci).
	 */
	public void individualBenefitShare(){
		Iterator<Person> ic = normativeDependence.keySet().iterator();
		double share=0;
		while(ic.hasNext()){
			Person person = ic.next();
			double choice = person.getComply();
			share+=choice;
		} 
		this.individualShare=share/(double)normativeDependence.size();	
	}
	
	/**
	 * ICi=(1/n)*Sum(((wij+2*Ij)*cj)/3), j!=i.
	 * @param p a person checking compliance with other agents
	 * @return pressure to comply
	 */
	public double influenceCompliance(Person p){
		Network network = (Network)RunState.getInstance().getMasterContext().getProjection("Population");
		Iterator<RelationshipEdge> ic = network.getOutEdges(p).iterator();
		double i=0;
		double result=0;
		
		while(ic.hasNext()){
			RelationshipEdge<Person> re = ic.next();
			Person person = re.getTarget();
			double weight=re.getWeight();
			double compliance=person.getComply();
			double choice=((weight+(2.0*person.getImportance()))*compliance)/3.0;		
			result+=choice;
			i++;
		}
		result=result/i;
		
		return result;
	}
	
	/**
	 * IAik=(1/n)*Sum(((wij+2*Ij)*ajk)/3), j!=i, k.
	 * @param person1 person evaluating approval of someone
	 * @param person2 person who is being evaluated
	 * @return approval of a peer
	 */
	public double influenceApproval(Person person1, Person person2){
		Network network = (Network)RunState.getInstance().getMasterContext().getProjection("Population");
		Iterator<RelationshipEdge> ic = network.getOutEdges(person1).iterator();
		double result=0;
		
		double i=0;
		while(ic.hasNext()){
			RelationshipEdge<Person> re=ic.next();
			Person person = re.getTarget();
			double weight=re.getWeight();
			if(person==person2)
				continue;	
			RelationshipEdge re2=(RelationshipEdge)network.getEdge(person,person2);
			double approval=re2.getApproval();
			result+=((weight+(person.getImportance()*2.0))*approval)/3.0;
			i++;
		}
		result=result/i;
		
		return result;
	}
	
	/**
	 * Ai=(1/N)*Sum(aji*wij), i!=j.
	 * @param p a person determining approval by peers
	 * @return average of approval by peers
	 */
	public double approvalByPeers(Person p){
		Network network = (Network)RunState.getInstance().getMasterContext().getProjection("Population");
		Iterator<RelationshipEdge> ic = network.getOutEdges(p).iterator();
		double result=0;
		
		double i=0;
		while(ic.hasNext()){
			RelationshipEdge<Person> re = ic.next();
			Person person = re.getTarget();
			double weight=re.getWeight();
			RelationshipEdge re2=(RelationshipEdge) network.getEdge(person,p);
			double approval=re2.getApproval();	
			result+=weight*approval;
			i++;
		}
		result=result/i;
		
		return result;
	}
	
	/**
	 * CSi=(1-Di)*S+Di*ci-ei.
	 * @param p person evaluating satisfaction with accepting 
	 * choice
	 * @return how satisfactory the previous choice was
	 */
	public double choiceSatisfaction(Person p){
		double satisfaction=0;
		
		double affectiveD=affectiveDependence.get(p);
		double wCost=p.getDecisionCost();

		satisfaction=((1.0-affectiveD)*this.individualShare)+(affectiveD*p.getComply())
			-wCost;
	
		if(satisfaction>1)
			satisfaction=1.0;
		if(satisfaction<-1)
			satisfaction=-1.0;
		
		return satisfaction;
	}
	
	/**
	 * ASij=(1-Di)*Aj+Di*aji*wij.
	 * @param person1
	 * @param person2
	 * @return satisfaction with approval of the other agent
	 */
	public double approvalSatisfaction(Person person1, Person person2){
		double approve=0;
		
		Network network = (Network)RunState.getInstance().getMasterContext().getProjection("Population");
		double affectiveD=affectiveDependence.get(person1);
		RelationshipEdge re = (RelationshipEdge) network.getEdge(person2, person1);
		double approvalOther=re.getApproval();
		RelationshipEdge re2 = (RelationshipEdge)network.getEdge(person1,person2);
		double weight=re2.getWeight();
		
		approve=((1.0-affectiveD)*approvalByPeers(person2))+(affectiveD*approvalOther*weight);
		
		if(approve>1)
			approve=1.0;
		if(approve<-1)
			approve=-1.0;
		
		return approve;
	}
	
	/**
	 * Delta CBi=(CSi+T(CBi))/2.
	 * @param p a person evaluating their bias to behave
	 * @return the bias to behave
	 */
	public double complianceBias(Person p){
		double cBias=0d;
		
		double satisfaction=choiceSatisfaction(p);
		double bias;
		
		if(!this.complianceBias.containsKey(p)) {
			bias=0;
		}
		else
			bias=this.complianceBias.get(p); 
		
		double linearImpact=linearImpact(bias);
		cBias=(satisfaction+linearImpact)/2.0;
		this.complianceBias.put(p, cBias); //update compliance bias
		
		return cBias;
	}
	
	/**
	 * Delta ABij=(ASij+T(ABij))/2, j!=i
	 * @param person1 person evaluating approval
	 * @param person2 person being evaluated for approval
	 * @return the approval bias
	 */
	public double approvalBias(Person person1, Person person2){
		Network network = (Network)RunState.getInstance().getMasterContext().getProjection("Population");
		RelationshipEdge re = (RelationshipEdge) network.getEdge(person1,person2);
		
		double aBias=0;
	
		double approvalS=approvalSatisfaction(person1,person2);
		double hBias;
		if(!this.approvalBias.containsKey(re)) {
			hBias=0;;
		}
		else
			hBias=this.approvalBias.get(re);
		
		double linearLimiting=linearImpact(hBias);
		aBias=(approvalS+linearLimiting)/2.0;
		
		this.approvalBias.put(re, aBias);
		
		return aBias;
	}
	
	/**
	 * if x>=0 then T=(1+x)/2, else T=(x-1)/2
	 * @param value input to adjust
	 * @return the adjusted linear value
	 */
	public double linearImpact(double value){
		double newValue=0;
		
		if(value>.1)
			newValue=(value+1.0)/2.0;
		if(value<-.1)
			newValue=(value-1.0)/2.0;
		
		return newValue;
	}
	
	/**
	 * PCi=(1-di)*CBi+di*ICi, while i!=Alpha.
	 * @param p a person who is determining their propensity to comply
	 */
	public double propensityToComply(Person p){
		double propensity=1;
		
		if(p.getAgentType()!=AgentType.Alpha){
			double nDependence=this.normativeDependence.get(p);
			double complianceBias=complianceBias(p);
			double influenceCompliance=influenceCompliance(p);
			propensity=((1-nDependence)*complianceBias)
				+(nDependence*influenceCompliance);
		}
		
		return propensity;
	}
	
	/**
	 * PAij=(1-Di)*ABij+Di*IAij.
	 * @param person1 person making approval choice
	 * @param person2 person being evaluated
	 * @return propensity to approve of a second agent (i.e. person2)
	 */
	public double propensityToApprove(Person person1,Person person2){
		double propensity=0;
		
		double nDepedence=this.normativeDependence.get(person1);
		double approvalBias=approvalBias(person1,person2);
		double influenceApproval=influenceApproval(person1,person2);
		propensity=((1.0-nDepedence)*approvalBias)+(nDepedence*influenceApproval);
		
		return propensity;
	}
		
	/**
	 * Delta wij=1/4(PCi*cj+T(wij)+aji*Di+Sum((PAij*akj)/n)), i!=j,k.
	 * @param person1 person evaluating weight
	 * @param person2 person being evaluated by person 1
	 * @return a weight change to the relationship edge
	 */
	public double weightChange(Person person1, Person person2,double propensityToChoose){
		Network network = (Network)RunState.getInstance().getMasterContext().getProjection("Population");
		
		double propensityToApprove=propensityToApprove(person1,person2);
		Iterator<RelationshipEdge> ic = network.getOutEdges(person1).iterator();
		double sumApproval=0;
		while(ic.hasNext()){
			RelationshipEdge<Person> re = ic.next();
			Person person = re.getTarget();
			if(person==person2)
				continue;
			RelationshipEdge re2 = (RelationshipEdge) network.getEdge(person,person2);	
			double approval = re2.getApproval();
			sumApproval+=(propensityToApprove*approval)/(double)(normativeDependence.keySet().size()-2);
		} 
		
		RelationshipEdge re = (RelationshipEdge) network.getEdge(person1,person2);
		RelationshipEdge re2 = (RelationshipEdge) network.getEdge(person2,person1);
     	
		double historyW=re.getWeight();
		double comply=person2.getComply();
		double otherA=re2.getApproval();
		double wWeight=linearImpact(historyW);
		
		double factor=(propensityToChoose*comply)+wWeight+
			(this.affectiveDependence.get(person1)*otherA)+sumApproval;
		double weightChange=factor/4.0d;
		
		re.setWeight(weightChange);
		updateApprovalState(person1,person2,propensityToApprove);
		
		return weightChange;
	}
	
	/**
	 * Method to adjust approval state between two agents based on PAij.
	 * @param person1 the person evaluating
	 * @param person2 the person evaluated
	 * @param approveP the approval value
	 */
	public void updateApprovalState(Person person1, Person person2,double approveP){
		Network network = (Network)RunState.getInstance().getMasterContext().getProjection("Population");
		
		double[] propensity=new double[]{approveP};
		double[] responseP=determineResponse(propensity);
		RelationshipEdge<Person> re = (RelationshipEdge) network.getEdge(person1, person2);
		re.setApproval(responseP[0]);
	}
	
	/**
	 * Adjustment method to modify propensity with values closer to extreme always 
	 * resulting in a deterministic result, but those close to zero are more random 
	 * and evaluated against a uniform distribution of probabilities.
	 * The logic is if(propensity>.2) then 1, if (propensity<-.2) then -1; 
	 * if (propensity<random) then -1, else 1.0.
	 * @param propensity the propensity array to modify
	 * @return an array with propensity
	 */
	public double[] determineResponse(double[] propensity){
		double[] responseP = new double[propensity.length];
		
		if(uniformPropensity==null)
			uniformPropensity = RandomHelper.createUniform(-1, 1);
		
		for(int i=0; i < propensity.length; i++){
			double randomP=uniformPropensity.nextDouble();
			double answer=0;
			if(propensity[i]>.2)
				answer=1.0;
			else if(propensity[i]<-.2)
				answer=-1.0;
			else {
				if(propensity[i]<randomP)
					answer=-1.0;
				else 
					answer=1.0;
			}
			responseP[i]=answer;
		}
		
		return responseP;
	}
	
	/**
	 * Method adjusts choice to agree or not based on PCi.
	 * @param person1 the person choosing
	 * @param propensity the propensity value
	 */
	public void updateChoicePropensity(Person person1,double propensity){
		double[] responseI={propensity};
		double[] response=determineResponse(responseI);
		person1.setComply((int)response[0]);
	} 
}
