package uaa.feral.social.waterUse;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.PriorityQueue;

import repast.simphony.engine.environment.RunEnvironment;
import repast.simphony.engine.environment.RunState;
import repast.simphony.engine.schedule.IAction;
import repast.simphony.engine.schedule.ISchedule;
import repast.simphony.engine.schedule.ScheduleParameters;
import repast.simphony.parameter.Parameters;
import repast.simphony.random.RandomHelper;

import uaa.feral.context.FeralContextBuilder;
import uaa.feral.context.OutputContext;
import uaa.feral.physical.HydrologicalSetting;
import uaa.feral.physical.HydrologicalUnit;
import uaa.feral.social.Community;
import uaa.feral.social.Person;
import uaa.feral.social.data.AppliedSocialModelData;
import uaa.feral.util.SpatialHydrologyCalculations;
import uaa.feral.util.Statistics;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.GeometryFactory;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.Point;

/**
 * Social distancing model based on UAA-Northwestern water use model by Daniel Stouffer.
 * @author 
 */
public class SocialDistancing extends WaterUseBehaviors{

	Point homeCoordinate;
	List<HydrologicalUnit> waterSourceContexts;
	List<HydrologicalUnit> waterSourcesSorted;
	List<Person> people;
	Map<HydrologicalUnit, Double> waterQuantity;
	Map<Integer,Person>ageCategory;
	
	boolean isMunicipalXCor;
	boolean isMunicipalyCor;
	boolean municipalUse=false;
	boolean start=true;
	
	int historyLength;
	int min_max_Length;
	
	double qualityThreshold;
	double weightFactor;
	double acceptableDistanceLevel;
	
	Map<Person,Boolean> municipalPreference; //preference for municipal water
	Map<Person,int[]>sourceVisits; //number of visits to sources
	Map<Person,double[]>sourceWeights; //weights for a given source
	Map<Person,double[]>sourcePreferences; //probability of selecting source
	Map<Person,Map<HydrologicalUnit,double[]>>quantityHistory; //given historical values of water use history
	Map<Person,Map<HydrologicalUnit,double[]>>quantityMinimum;
	Map<Person,Map<HydrologicalUnit,double[]>>quantityMaximum;
	Map<Person,double[]>quantityBelief;
	Map<Person,double[]>biasedQuantityBelief;
	Map<Person,Double>globalQuantityBelief;
	Map<Person,double[]>quantityChangeBelief;
	Map<Person,double[]>biasedQuantityChangeBelief;
	Map<Person,Double>globalQuantityChangeBelief;
	Map<Person,HydrologicalUnit> currentWaterSource;
	private int waterSize;
	private Community community;
	private double mean_min;
	private double meanMax;
	private double endTick;
	private OutputContext outputContext;
	private int tickNow;
	private AppliedSocialModelData socialDistancingModelData;
	private double year;
	
	/**
	 * Class to run model continuously
	 * @author Mark Altaweel
	 */
	private class RunWaterUse implements IAction{
		public void execute() {
			runModel();
		}
	}
	
	/**
	 * Main constructor for object
	 * @param community a community that will implement model
	 */
	public SocialDistancing(Community community){
		this.community=community;
	}
	
	/**
	 * Method to make the hydrological quantity noisy
	 * @param hu the hydrological unit used by an agent
	 * @return a double quantity value
	 */
	public double noisyQuantity (HydrologicalUnit hu) {
		double noisy=0d;
		if(hu.isMunicipalSource())
			return hu.getVolume();
		
		double wBaseline = hu.calculateDailyVolumeFromDischarge();
		
		double wAmplitude = hu.getAmplitude()*wBaseline;
		
		noisy=RandomHelper.createNormal(wBaseline, wAmplitude/2.5).nextDouble();
		
		return noisy;
	}
	
	/**
	 * Method degrades water quantity  by .05 percent
	 */
	public void degradeWaterQuantity(){
			Iterator<HydrologicalUnit> ic = waterSourcesSorted.iterator();
			while(ic.hasNext()){
				HydrologicalUnit hu = ic.next();
				double volume=hu.getVolume();
				if(!hu.isMunicipalSource()) {
					volume=hu.getDischarge()*.995;
					hu.setDischarge(volume);
				} 
				else{
					volume=volume*.995;
					hu.setVolume(volume);
				}
				double minV1=Math.min(.05,Double.MAX_VALUE-volume);
				double minV2=Math.min(minV1, volume-0);
				hu.setAmplitude(RandomHelper.nextDoubleFromTo(0,minV2));
				waterQuantity.put(hu, volume);
			}
	}
	
	/**
	 * Method degrades municipal quality
	 */
	public void degradeWaterQuality(){
		Iterator<HydrologicalUnit> ic = waterSourcesSorted.iterator();
		while(ic.hasNext()){
			HydrologicalUnit hu = ic.next();
			hu.setQuality(hu.getQuality()*.99995);
		}	
	}
	
	/**
	 * Method calls methods to degrade quality(in municipal systems) or quantity (in non-municipal systems)
	 */
	private void toWaterCycle(){
		if(degradeQualityMode)
			degradeWaterQuality();
	}
	
	/**
	 * Method calls method to update water quality/quantity
	 */
	private void updateWorld(){
		toWaterCycle();
	}
	
	/**
	 * Method launched when model first starts. It initializes data and sets the model for further functions
	 */
	public void initializeWater(){
		
		//start this at step 0	
		homeCoordinate=community.getLocation();
		
		PriorityQueue<HydrologicalUnit> unsettled = new PriorityQueue<HydrologicalUnit>( waterSourceContexts.size(),
			new Comparator<HydrologicalUnit>() {
			public int compare(HydrologicalUnit o1, HydrologicalUnit o2) {
				double distance=SpatialHydrologyCalculations.LatLongDistance(homeCoordinate.getCoordinate(),o1.getGeometry().getCentroid().getCoordinate());
				double distance2=SpatialHydrologyCalculations.LatLongDistance(homeCoordinate.getCoordinate(),o2.getGeometry().getCentroid().getCoordinate());	
				return distance < distance2 ? -1 : distance==distance2 ? 0 : 1;
			}
		}); 
		int i=0;
		Iterator<HydrologicalUnit>ic = waterSourceContexts.iterator();
		while(ic.hasNext()) {
			HydrologicalUnit hu = ic.next();
			unsettled.add(hu);
		}
		
		int size=unsettled.size();
//		List<String> tempName=new ArrayList<String>();
		while(i < size) {
			HydrologicalUnit hu=unsettled.poll();
			double distance=SpatialHydrologyCalculations.LatLongDistance(homeCoordinate.getCoordinate(),
					hu.getGeometry().getCentroid().getCoordinate());
			double latitudeDifference=homeCoordinate.getCoordinate().x-hu.getGeometry().getCentroid().getX();
			i++;
			hu.setDistanceToTown(distance);
			if(hu==null || distance>this.acceptableDistanceLevel ||latitudeDifference<0)
				continue;
//			if(!tempName.contains(hu.getName()))
//				tempName.add(hu.getName());
//			else
//				continue;
	
			if(hu.getVolume()>0)
				waterQuantity.put(hu,hu.getVolume());
			else
				waterQuantity.put(hu, hu.getDischarge());
			waterSourcesSorted.add(hu);
			hu.setIdentification(waterSourcesSorted.size()-1);
		}
		waterSize=waterSourcesSorted.size();
	}
	
	/**
	 * Method that initializes data values and data containers used
	 */
	public void initialValues(){
		//start this at 0
		waterSourcesSorted=new ArrayList<HydrologicalUnit>();
		municipalPreference=new HashMap<Person,Boolean>();
		sourceVisits=new HashMap<Person,int[]>();
		sourceWeights=new HashMap<Person,double[]>();
		sourcePreferences=new HashMap<Person,double[]>(); 
		quantityHistory=new HashMap<Person,Map<HydrologicalUnit,double[]>>();
		quantityMinimum=new HashMap<Person,Map<HydrologicalUnit,double[]>>();
		quantityMaximum=new HashMap<Person,Map<HydrologicalUnit,double[]>>();
		quantityBelief=new HashMap<Person,double[]>();
		biasedQuantityBelief=new HashMap<Person,double[]>();
		globalQuantityBelief=new HashMap<Person,Double>(); 
		quantityChangeBelief=new HashMap<Person,double[]>();
		biasedQuantityChangeBelief=new HashMap<Person,double[]>();
		globalQuantityChangeBelief=new HashMap<Person,Double>();
		currentWaterSource=new HashMap<Person,HydrologicalUnit>();
		waterQuantity=new HashMap<HydrologicalUnit,Double>();
		historyLength=8;
		min_max_Length=5;
		qualityThreshold=.75d;
		weightFactor=.1d;
	}
	
	/**
	 * Method to reset data values to different quantities
	 */
	public void resetValues(){
		Iterator<Person> ic = people.iterator();
		while(ic.hasNext()){
			Person p = ic.next();
			sourceVisits.put(p, createValues(waterSize+1,0));
			quantityBelief.put(p, createValues(waterSize,RandomHelper.nextDouble()));
			quantityChangeBelief.put(p, createValues(waterSize,.5d));
			double[] sWeights=sourceWeights.get(p);
			double[] nSWeights=new double[sWeights.length+1];
			double[] sPreference=sourcePreferences.get(p);
			double sum=0d;
			for(int i=0; i < sWeights.length; i++){
				sum+=sWeights[i];
			}
			for(int i=0; i < nSWeights.length;i++){
				if(i<sWeights.length){
					nSWeights[i]=sWeights[i];
					sPreference[i]=(sPreference[i]/sum);
				}
				else{
					nSWeights[i]=(sum)/(waterSize+1)*(1-1/waterSize+1);
				}
			}
			sourceWeights.put(p, nSWeights);
			organizeWaterForPeople(p);
		}
	}
	
	/**
	 * Method associates a person agent with different water use quantities and data holders
	 * @param person a person agent
	 */
	public void organizeWaterForPeople(Person person){
		Iterator<HydrologicalUnit> ic = waterSourcesSorted.iterator();
		while(ic.hasNext()){
			HydrologicalUnit unit = ic.next();
			Map<HydrologicalUnit, double[]>  qH= new HashMap<HydrologicalUnit, double[]>();
			Map<HydrologicalUnit, double[]>  qM= new HashMap<HydrologicalUnit, double[]>();
			Map<HydrologicalUnit, double[]>  qMi= new HashMap<HydrologicalUnit, double[]>();
			qH.put(unit,null);
			qM.put(unit, null);
			qMi.put(unit, null);
			quantityHistory.put(person,qH);
			quantityMinimum.put(person, qM);
			quantityMaximum.put(person, qMi);
		}
	}
	
	/**
	 * Method that calls the methods that move an agent to a given water source location.
	 * This method has been changed from Daniel's original version
	 * @param p a person agent
	 */
	public void moveToWaterSource(Person p) {
		if(currentWaterSource.get(p)==null || ((HydrologicalUnit)currentWaterSource.get(p)).getName().contains("Municipal"))
			moveToMunicipalWaterSource(p);
		else
			moveToNaturalWaterSource(p);
	}
	
	/**
	 * Method that calculates water source preferences
	 * @param hu a hydrological unit object
	 * @param p a person agent
	 */
	public void bayesianPreference(HydrologicalUnit hu,Person p){
		
		double quality=hu.getQuality();
		double[] sourcePref=sourcePreferences.get(p);
		double prefValue=0d;
		if(hu!=null)
			prefValue=sourcePref[hu.getIdentification()];	
		
		double denomPart=quality*prefValue;
		
		double sumSourcePrefValue=0d;
		for (int i=0; i < sourcePref.length;i++){
			if(i!=hu.getIdentification())
				sumSourcePrefValue=+(1-quality)/(sourcePref.length-1)*sourcePref[i];
		}
		double denom=denomPart+sumSourcePrefValue;
		double[] previousPref=sourcePref;
		
		sourcePref=new double[sourcePref.length];
		List<Double>tempList=new ArrayList<Double>();
		for(int i=previousPref.length-1; i>-1; i--){
			double addItem=0d;
			if(hu.getIdentification()==i)
				addItem=(quality*previousPref[i]/denom);
			else
				addItem=((1-quality)/(previousPref.length-1)/denom);
			
			tempList.add(addItem);
		}
		
		for(int i=0; i < tempList.size();i++){
			sourcePref[i]=tempList.get(i);
		}
		
		sourcePreferences.put(p, sourcePref);
	}
	
	/**
	 * Method to give weights to sources visited
	 * @param p a person agent
	 */
	public void visitWeightedPreference(Person p){
//		double[] sWeights=sourceWeights.get(p);
		int[] sWeights=sourceVisits.get(p);
		double[] sPreferences=sourcePreferences.get(p);
		HydrologicalUnit hu = currentWaterSource.get(p);
//		sWeights[hu.getIdentification()]=hu.getQuality()+sWeights[hu.getIdentification()];
//		double sumValue=Statistics.computeSum(sWeights);
		double sumValue=Statistics.computeSum(sWeights);
		double sumPref=Statistics.computeSum(sPreferences);
		for(int i=0; i < sPreferences.length;i++){
//			sPreferences[i]=sWeights[i]/sumValue;
			sPreferences[i]=(sWeights[i]+sPreferences[i])/(sumValue+sumPref);
		}
		p.setSourcePreference(sPreferences[hu.getIdentification()]);
	}
	
	/**
	 * Method to move to a municipal water source
	 * @param p a person agent
	 */
	public void moveToMunicipalWaterSource(Person p){
		Community community=p.getCommunity();
		p.moveTo(community.getLocation());
	}
	
	/**
	 * Method to move to non-municipal water sources
	 * @param p a person agent
	 */
	public void moveToNaturalWaterSource(Person p){
		HydrologicalUnit hu = currentWaterSource.get(p);
		if(RunEnvironment.getInstance().getCurrentSchedule().getTickCount() % 3==0){ //purely for visual purposes, no real meaning to this
			  Coordinate coords[] = ((MultiLineString)hu.getGeometry()).getCoordinates();
			  Coordinate c=coords[RandomHelper.nextIntFromTo(0, coords.length-1)];
			  Coordinate coord=(Coordinate) c.clone();
			  GeometryFactory f = new GeometryFactory();
			  p.moveTo(f.createPoint(coord));
		}
		else if(RunEnvironment.getInstance().getCurrentSchedule().getTickCount() % 2==0){
			  Coordinate coords[] = ((MultiLineString)hu.getGeometry()).getCoordinates();
			  Coordinate c=coords[RandomHelper.nextIntFromTo(0, coords.length-1)];
			  Coordinate coord=(Coordinate) c.clone();
			  double x=p.getLocation().getX();			  
			  double y=p.getLocation().getY();
			  coord.x=(double)(coord.x+x)/2;
			  coord.y=(double)(coord.y+y)/2;
        	  GeometryFactory f = new GeometryFactory();
        	  p.moveTo(f.createPoint(coord));
		}
		else
			p.moveTo(community.getLocation());
	}
	
	/**
	 * Method used in intializing model values associated with agents
	 * @param size the size of an array
	 * @param initValue the initial values in the array
	 * @return a int array
	 */
	public int[] createValues(int size,int initValue){
		int[] values=new int[size];
		for(int i=0; i < size; i++){
			values[i]=initValue;
		}
		return values;
	}
	
	/**
	 * Method used in initializing model values associated with agents
	 * @param size the size of the array
	 * @param initValue the initial values in the arrray
	 * @return a double array
	 */
	public double[] createValues(int size,double initValue){
		double[] values=new double[size];
		for(int i=0; i < size; i++){
			values[i]=initValue;
		}
		return values;
	}
	
	public double getWeightFactor() {
		return weightFactor;
	}


	public void setWeightFactor(double weightFactor) {
		this.weightFactor = weightFactor;
	}
	
	/**
	 * Method to run the model that is called by the community
	 */
	public void runModel(){
		this.tickNow=(int)RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
		if(outputContext==null)
			outputContext=(OutputContext) RunState.getInstance().getMasterContext().getSubContext("output");
		updatePeople();
		if(start){
			setupModelData(scenarioData.getSocialModelDataFile());
			this.ageCategory=new HashMap<Integer,Person>();
			this.waterSourceContexts=new ArrayList<HydrologicalUnit>();
			Iterator<Object> ic=RunState.getInstance().getMasterContext().getSubContexts().iterator();
			while(ic.hasNext()){
				Object c = ic.next();
				if(c instanceof HydrologicalSetting)
					determineSize((HydrologicalSetting)c);
			}
			initialValues();
			initializeWater();
			Parameters p = RunEnvironment.getInstance().getParameters();
			double numberYears=(Integer)p.getValue("numberOfYears");
			int daysPerTick=(Integer)p.getValue("daysPerTick");
			this.endTick=((double) 365.25/daysPerTick)*numberYears;
			this.year=(double)365.25/daysPerTick;
			start=false;
		}
		updateWorld();
		askAgents();
		outputContext.recordWaterUseOutput();
		if(tickNow==(int)endTick){
			outputContext.compileOutput();
			return;
		}
		
		ISchedule schedule2 = repast.simphony.engine.environment.RunEnvironment.getInstance()
      	.getCurrentSchedule();
			schedule2.schedule(ScheduleParameters.createOneTime(tickNow+1,0),new RunWaterUse());
	}
	
	/**
	 * Method updates the people from the community
	 */
	public void updatePeople(){
		if(people==null)
			people=new ArrayList<Person>();
		Iterator<Person> ic = people.iterator();
		while(ic.hasNext()){
			Person p = ic.next();
			if(this.largeSimulation) {
				if(ageCategory.containsValue(p))
					returnHome(p);
			}
			else
				returnHome(p);
		}
		people.clear();
		people.addAll(community);
	}
	
	/**
	 * Method determines the size or number of water sources from a given system
	 * @param hs a hydrological system
	 */
	public void determineSize(HydrologicalSetting hs){
		Iterator<HydrologicalUnit>hi=hs.getObjects(HydrologicalUnit.class).iterator();
		while(hi.hasNext()){
			HydrologicalUnit h = hi.next();
			
			if(h.isMunicipalSource() && h.getName().contains((String)community.getId())) {
				waterSourceContexts.add(h);
				this.municipalUse=true;
			}
			
			else if(h.isMunicipalSource())
				continue;
			
			if(!h.isMunicipalSource())
				waterSourceContexts.add(h);	        		
		}
	}
	
	/**
	 * Method moves the agents home (i.e the location of their community)
	 * @param p
	 */
	public void returnHome(Person p){
		moveToMunicipalWaterSource(p);
	}

	/**
	 * Method that allows agents to get water and update values associated with water (quantity and quality) and perceptions of water
	 */
	public void askAgents(){
		ageCategory.clear();
		for(int i=0; i < people.size(); i++) {
			Person p = people.get(i);
			if(!p.isAlive()) {
				community.remove(p);
				continue;
			}
			if(p.getAge()>=15 && p.getAge()< 85 && p.getId()%2==0) {
				if(this.largeSimulation) {		
					if(ageCategory.containsKey(p.getAge()) && !ageCategory.containsValue(p))
						continue; 
					ageCategory.put(p.getAge(), p);
				}
				checkPerson(p);
				HydrologicalUnit hu=selectWaterSource(p);
				currentWaterSource.put(p,hu);
				moveToWaterSource(p);
				collectWater(p,hu);
				source(p);
				if(!outputContext.contains(p))
					outputContext.add(p);
				if(!outputContext.contains(hu))
					outputContext.add(hu);
	//			bayesianPreference(hu,p);
			}
			else
				continue;
		}
	}
	
	/**
	 * Method that checks to see if the person agent has previously been used in the model
	 * @param p a person agent
	 */
	public void checkPerson(Person p){
		if(!sourceVisits.containsKey(p)){
			sourceVisits.put(p, createValues(waterSize,0));
			sourceWeights.put(p,createValues(waterSize,1d));
			sourcePreferences.put(p,weightWaterByDistance(p,waterSize));
			quantityBelief.put(p, createValues(waterSize,0d));
			biasedQuantityBelief.put(p, createValues(waterSize,0d));
			quantityChangeBelief.put(p, createValues(waterSize,0d));
			biasedQuantityChangeBelief.put(p, createValues(waterSize,0d));
			organizeWaterForPeople(p);
		}
	}
	
	/**
	 * Weight water sources by distance
	 * @param size number of water sources
	 * @return weighted values in array
	 */
	public double[] weightWaterByDistance(Person person,int size){
		double[] weights=new double[size];
			
		double sum=0;
		for(int i=0; i < waterSourcesSorted.size(); i++){
			HydrologicalUnit hu = waterSourcesSorted.get(i);	
			double w=0;
			w=1/Math.pow(hu.getDistanceToTown(),2.2);
			if(hu.getDistanceToTown()==0){
				w=0;
			}
				
			weights[i]=w;
			sum+=weights[i];
		}
		if(weights[0]==0){
			if(person.getAge()<=39+(this.tickNow/this.year)) {
				weights[0]=sum*this.socialDistancingModelData.getAgeWeightedDecisions().get("young");
			}
			else if(person.getAge()<=59+(this.tickNow/this.year)) {
				weights[0]=sum*this.socialDistancingModelData.getAgeWeightedDecisions().get("middle"); 
			}
			else {
				System.out.println(this.tickNow);
				weights[0]=sum*this.socialDistancingModelData.getAgeWeightedDecisions().get("old");
			}
			
			for(int i=1; i < weights.length;i++){
				weights[i]=(1-(weights[0]/sum))*weights[i];
			}
		}
		for(int i=0; i < weights.length;i++){
			weights[i]=weights[i]/sum;
		}
		return weights;
	}
	
	/**
	 * Method that calls the method for an agent to search for a water source
	 * @param p a person agent
	 */
	public void source(Person p){
		if(updatePreferences)
			updateSourcePreferences(p);	
	}
	
	/**
	 * Method updates an agents preference for water sources
	 * @param p a person agent
	 */
	public void updateSourcePreferences(Person p){
		HydrologicalUnit hu = currentWaterSource.get(p);
		int visits[]=sourceVisits.get(p);
		visits[hu.getIdentification()]+=1;
		visitWeightedPreference(p);
		hu.setTimesVisited(hu.getTimesVisited()+1);
		p.setSourceVisited(hu.getIdentification());
	}
	
	/**
	 * Method that selects a hydrological source for an agent to use
	 * @param person a person agent
	 * @return a hydrological unit selected by the agent
	 */
	public HydrologicalUnit selectWaterSource(Person person){
		HydrologicalUnit unit=null;
		double s=0d;
	
		double r=RandomHelper.nextDouble();
		int i=-1;
		double[] sourcePref=sourcePreferences.get(person);
		while(s<r){
			i++;
			try{
				s+=sourcePref[i];
			}
			catch(ArrayIndexOutOfBoundsException e){
				i=waterSourcesSorted.size()-1;
				break;
			}
		}
		
		unit=waterSourcesSorted.get(i);
		
		return unit;
	}
	
	/**
	 * Method that calls a method to update quantity extracted by an agent
	 * @param person a person agent
	 * @param hu a hydroloigical unit source
	 */
	public void collectWater(Person person,HydrologicalUnit hu){
		updateQuantityRecord(hu,person,noisyQuantity(hu));
	}
	
	/**
	 * Method that calls methods that updates different quantity related values (history, 
	 * minimum, maximum, quantity belief, and quantity change belief)
	 * @param hu the hydrological unit
	 * @param person a person agent
	 * @param hydrological unit quantity to update
	 */
	public void updateQuantityRecord(HydrologicalUnit hu,Person person,double quantity){
		updateQuantityHistory(hu,person,quantity);
		updateQuantityMinimum(hu,person,quantity);	
		updateQuantityMaximum(hu,person,quantity);
		
		updateQuantityBelief(person,hu);
		updateQuantityChangeBelief(person,hu);
	}
	
	/**
	 * Method that updates the history kept by an agent about quantities (latest historical values based on length of historical memory)
	 * @param hu a hydrological unit
	 * @param person person agent
	 * @param quantity the quantity to update in memory
	 */
	public void updateQuantityHistory(HydrologicalUnit hu,Person person, double quantity){
		Map<HydrologicalUnit,double[]>inputMap=quantityHistory.get(person);
		if(inputMap==null) {
			inputMap=new HashMap<HydrologicalUnit,double[]>();
			quantityHistory.put(person,inputMap);
		}
		if(inputMap.get(hu)==null){
			double[] q = {quantity};
			inputMap.put(hu, q);
			return;
		}
			
		double[] qHistory = inputMap.get(hu);
	
		int length=0;
		if(qHistory.length<historyLength){
			length=qHistory.length+1;
		}
		else
			length=historyLength;
		
		double[] tempA=new double[length];
		for(int i=0; i < tempA.length;i++){
			if(i==0)
				tempA[i]=quantity;
			else
				tempA[i]=qHistory[i-1];
			}
		qHistory=tempA;
		inputMap.put(hu, qHistory);
	}
	
	/**
	 * Method that checks to see if the water quantity is among the minimum amounts kept in memory by an agent
	 * @param hu a hydrological unit
	 * @param person a person agent
	 * @param waterToCollect amount of water collected
	 */
	public void updateQuantityMinimum(HydrologicalUnit hu,Person person,double waterToCollect){
		Map<HydrologicalUnit,double[]> inputMap=quantityMinimum.get(person);
		if(inputMap==null) {
			inputMap=new HashMap<HydrologicalUnit,double[]>();
			quantityMinimum.put(person,inputMap);
		}
		if(inputMap.get(hu)==null){
			double[] q = {waterToCollect};
			inputMap.put(hu, q);
			return;
		}
		double[] quantityM = inputMap.get(hu);
		
		
		List<Double>sortList=new ArrayList<Double>();
		
		for(int i=0; i < quantityM.length;  i++){
			sortList.add(new Double(quantityM[i]));
		}
		sortList.add(new Double(waterToCollect));
		
		int length=0;
		if(quantityM.length<min_max_Length) {
			length=quantityM.length+1;
		}
		else
			length=min_max_Length;
		
		double[] mHistory=new double[length];
		Collections.sort(sortList);
		
		for(int i=0; i < length; i++){
			mHistory[i]=sortList.get(i);
		}
		inputMap.put(hu,mHistory);
			
	}
	
	/**
	 * Method to see if the water collected is an amount among the record high amounts of water collected in agent memory
	 * @param hu the hydrological unit
	 * @param person a person agent
	 * @param waterToCollect the amount of water collected
	 */
	public void updateQuantityMaximum(HydrologicalUnit hu,Person person,double waterToCollect){
		Map<HydrologicalUnit,double[]> inputMap=quantityMaximum.get(person);
		if(inputMap==null) {
			inputMap=new HashMap<HydrologicalUnit,double[]>();
			quantityMaximum.put(person,inputMap);
		}
		if(inputMap.get(hu)==null){
			double[] q = {waterToCollect};
			inputMap.put(hu, q);
			return;
		}
		double[] quantityM = inputMap.get(hu);
		
		List<Double>sortedList=new ArrayList<Double>();
		for(int i=0; i < quantityM.length;  i++){
			sortedList.add(new Double(quantityM[i]));
		}
		
		sortedList.add(new Double(waterToCollect));
		int length=0;
		if(quantityM.length<min_max_Length)
			length=quantityM.length+1;
		else
			length=min_max_Length;
		
		double[] mHistory=new double[length];
		Collections.sort(sortedList);
		int i=sortedList.size()-1;
		int j=0;
		while(true){
			try{
				mHistory[j]=sortedList.get(i);
			}
			catch(IndexOutOfBoundsException e){
				break;
			}
			i--;
			j++;
		}
	
		inputMap.put(hu,mHistory);
			
	}
	
	/**
	 * Method used to weight certain agents
	 * @param timeSeries time series values
	 * @return a weight value
	 */
	public double ageWeighted(double[] timeSeries){
		//not currently used
		double sumN=0d;
		double sumD=0d;
		double tickCount=RunEnvironment.getInstance().getCurrentSchedule().getTickCount();
		for(int i=0; i < timeSeries.length; i++){
			sumN+=timeSeries[0]*(1/(tickCount-timeSeries[timeSeries.length-1]+1));
			sumD+=1/(tickCount-timeSeries[timeSeries.length-1]+1);
		}
		return sumN/sumD;
	}
	
	/**
	 * Method to weight preferences for certain water sources
	 * @param person an agent that will update his/her preference for water sources
	 * @param belief belief values associated with water source
	 * @return a double array of weighted water preferences
	 */
	public double[] preferenceWeighted(Person person,double[] belief){
		double j=0;
		double[] sourcePref=sourcePreferences.get(person);
		double[] weightedBelief=new double[sourcePref.length];
		for(int i=0; i < belief.length; i++){
			if(belief[i]==0)continue;
			j+=(double)belief[i]*sourcePref[i];
		}
		
		for(int i=0; i < belief.length; i++){			
			if(belief[i]==0)continue;
			weightedBelief[i]=(sourcePref[i] * (double)belief[i]) + ((1-sourcePref[i])*j);
		}
		
		return weightedBelief;
	}
	
	/**
	 * Method updates the belief of an agent on how much water there is for a given water source
	 * @param person a person agent
	 * @param hu a hydrological unit
	 */
	public void updateQuantityBelief(Person person,HydrologicalUnit hu){
		double sumV=0d;
		double sumM=0;
		double sumMi=0;
		
		double[]qBelief=quantityBelief.get(person);
		Iterator<Person> ic = quantityHistory.keySet().iterator();
		
		int size=0;
		while(ic.hasNext()){
			Person p = ic.next();
			if(!p.isAlive()){
				removeAgent(p);
				ic=quantityHistory.keySet().iterator();
				continue;
			}
				
			Map<HydrologicalUnit,double[]> item=quantityHistory.get(p);
			Map<HydrologicalUnit,double[]> item2=quantityMinimum.get(p);
			Map<HydrologicalUnit,double[]> item3=quantityMaximum.get(p);
			
			if(!item.containsKey(hu) || item.get(hu)==null)
				continue;
			
			double qH[] = item.get(hu);
			sumV+=qH[0];
		
			double qM[] = item2.get(hu);
			double qMi[] = item3.get(hu);
			sumM+=qM[0];
			sumMi+=qMi[0];
			
			size++;
		}
			
		double mean=(double)sumV/size;
		mean_min=(double)sumM/size;
		meanMax=(double)sumMi/size;
		qBelief[hu.getIdentification()]=mean;
		
		double[] bBelief=biasedQuantityBelief.get(person);
		bBelief=preferenceWeighted(person,qBelief);
		biasedQuantityBelief.put(person, bBelief);
		
		double sumQBelief=0d;
		double[] sourcePref=sourcePreferences.get(person);
		for(int i=0; i < qBelief.length;i++){
			sumQBelief+=qBelief[i]*sourcePref[i];
		}

		person.setQuantityBelief(qBelief[hu.getIdentification()]);  //quantity belief is the average of the sources' quantity 
		globalQuantityBelief.put(person, sumQBelief);
	}
	
	/**
	 * Method to remove agent from the models data
	 * @param p a person agent
	 */
	public void removeAgent(Person p){
		municipalPreference.remove(p); 
		sourceVisits.remove(p); 
		sourceWeights.remove(p);
		sourcePreferences.remove(p); 
		quantityHistory.remove(p); 
		quantityMinimum.remove(p);
		quantityMaximum.remove(p);
		quantityBelief.remove(p);
		biasedQuantityBelief.remove(p);
		globalQuantityBelief.remove(p);
		quantityChangeBelief.remove(p);
		biasedQuantityChangeBelief.remove(p);
		globalQuantityChangeBelief.remove(p);
		currentWaterSource.remove(p);		
	}
	
	/**
	 * Method that updates an agents belief about quantity changes in the water source used
	 * @param person a person agent
	 * @param hu a hydrological unit
	 */
	public void updateQuantityChangeBelief(Person person,HydrologicalUnit hu){
		double recent=quantityBelief.get(person)[hu.getIdentification()];

		double[] qCB=quantityChangeBelief.get(person);
		if(hu.isMunicipalSource())
			qCB[hu.getIdentification()]=1;
		else{
			if(meanMax==mean_min){
				qCB[hu.getIdentification()]=.5d;
			}
			else{
				qCB[hu.getIdentification()]=Math.abs((recent-mean_min)/(meanMax-mean_min));
			}
		}
		
		double[] bBelief=biasedQuantityChangeBelief.get(person);
		bBelief=preferenceWeighted(person,qCB);
		biasedQuantityChangeBelief.put(person, bBelief);
		
		double[] sourcePref=sourcePreferences.get(person);
		double sumVG=0d;
		for(int i=0;i<sourcePref.length;i++){
			sumVG+=sourcePref[i]*qCB[i];
		}
		
		person.setQuantityChangeBelief(qCB[hu.getIdentification()]);
		globalQuantityChangeBelief.put(person, sumVG);
	}

	public void setupModelData(String file){
		AppliedSocialModelData  sdmd = new AppliedSocialModelData();
		this.socialDistancingModelData=sdmd.setupModelData(FeralContextBuilder.PathToData+file);
		this.acceptableDistanceLevel=socialDistancingModelData.getTolerableDistanceToSource();
		this.largeSimulation=socialDistancingModelData.isLargeSim();
		this.municipalUse=socialDistancingModelData.isMunicipal();
	}
	
	public List<HydrologicalUnit> getWaterSourceContexts() {
		return waterSourceContexts;
	}

	public void setWaterSourceContexts(List<HydrologicalUnit> waterSourceContexts) {
		this.waterSourceContexts = waterSourceContexts;
	}

	public double getAcceptableDistanceLevel() {
		return acceptableDistanceLevel;
	}

	public void setAcceptableDistanceLevel(double acceptableDistanceLevel) {
		this.acceptableDistanceLevel = acceptableDistanceLevel;
	}
}
